"""Plotting functions for visualising circumplex data."""

from typing import Union, Tuple, List

import matplotlib as mpl
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import pearsonr, spearmanr

diag_lines_zorder = 1
diag_labels_zorder = 4
prim_lines_zorder = 2
data_zorder = 3
default_bw_adjust = 1.2
default_figsize = (5, 5)

simple_density = dict(thresh=0.5, levels=2, incl_outline=True, alpha=0.5)


def scatter(
    data=None,
    x="ISOPleasant",
    y="ISOEventful",
    title="Soundscape Scatter Plot",
    diagonal_lines=False,
    xlim=(-1, 1),
    ylim=(-1, 1),
    figsize=(5, 5),
    legend_loc="lower left",
    hue=None,
    style=None,
    s=20,
    palette="colorblind",
    hue_order=None,
    hue_norm=None,
    sizes=None,
    size_order=None,
    size_norm=None,
    markers=True,
    style_order=None,
    alpha=None,
    legend="auto",
    ax=None,
    **scatter_kws,
):
    """Plot ISOcoordinates as scatter points on a soundscape circumplex grid

    Makes use of seaborn.scatterplot. We have made all of the `seaborn.scatterplot` arguments available, but have also added or changed some specific
    options for circumplex plotting.

    Parameters
    ----------
    data : pd.DataFrame, np.ndarray, mapping or sequence
        Input data structure. Either a long-form collection of vectors that can be assigned to
        named variables or a wide-form dataset that will be internally reshaped.
    x : vector or key in `data`, optional
        column name for x variable, by default "ISOPleasant"
    y : vector or key in `data`, optional
        column name for y variable, by default "ISOEventful"
    title : str, optional
        Title to add to circumplex plot, by default "Soundscape Scatter Plot"
    diagonal_lines : bool, optional
        whether to include diagonal dimension labels (e.g. calm, etc.), by default False
    xlim, ylim : tuple, optional
        Limits of the circumplex plot, by default (-1, 1)
        It's recommended to set these such that the x and y axes have the same aspect
    figsize : tuple, optional
        Size of the figure to return if `ax` is None, by default (5, 5)
    legend_loc : str, optional
        relative location of legend, by default "lower left"
    palette : string, list, dict or matplotlib.colors.Colormap, optional
        Method for choosing the colors to use when mapping the hue semantic. String values are passed to
        seaborn.color_palette(). List or dict values imply categorical mapping, while a colormap object
        implies numeric mapping.
        by default colorblind
    s : int, optional
        size of scatter points, by default 10
    hue : vector or key in data, optional
        Grouping variable that will produce points with different colors. Can be either categorical or numeric,
        although color mapping will behave differently in latter case, by default None
    style : vector or key in data
        Grouping variable that will produce points with different markers. Can have a numeric dtype but will always
        be treated as categorical.
    hue_order : vector of strings
        Specify the order of processing and plotting for categorical levels of the `hue` semantic
    hue_norm : tuple or matplotlib.colors.Normalize
        Either a pair of values that set the normalization range in data units or an object that will map from data
        units into a [0, 1] interval. Usage implies numeric mapping.
    sizes : list, dict, or tuple
        An object that determines how sizes are chosen when `size` is used. It can always be a list of size values or
        a dict mapping levels of the `size` variable to sizes. When `size` is numeric, it can also be a tuple
        specifying the minimum and maximum size to use such that other values are normalized within this range.
    size_order : list
        Specified order for appearance of the `size` variable levels, otherwise they are determined from the data. Not
        relevant when the `size` variable is numeric.
    size_norm : tuple or Normalization object
        Normalization in data units for scaling plot objects when the `size` variable is numeric.
    markers : boolean, list, or dictionary
        Object determining how to draw the markers for different levels of the `style` variable. Setting to `True` will
        use default markers, or you can pass a list of markers or a dictionary mapping levels of the `style` variable
        to markers. Setting to `False` will draw marker-less lines. Markers are specified as in matplotlib.
    style_order : list
        Specified order for appearance of the `style` variable levels otherwise they are determined from the data. Not
        relevant when the `style` variable is numeric.
    alpha : float
        Proportional opacity of the points.
    legend : {"auto", "brief", "full" or False}, optional
        How to draw the legend. If “brief”, numeric hue and size variables will be represented with a sample of evenly
        spaced values. If “full”, every group will get an entry in the legend. If “auto”, choose between brief or full
        representation based on number of levels. If False, no legend data is added and no legend is drawn.
        by default, "auto"
    ax : matplotlib.axes.Axes, optional
        Pre-existing matplotlib axes for the plot, by default None
        If `None` call `matplotlib.pyplot.subplots` with `figsize` internally.

    Returns
    -------
    matplotlib.axes.Axes
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)

    s = sns.scatterplot(
        data=data,
        x=x,
        y=y,
        hue=hue,
        style=style,
        s=s,
        palette=palette,
        hue_order=hue_order,
        hue_norm=hue_norm,
        sizes=sizes,
        size_order=size_order,
        size_norm=size_norm,
        markers=markers,
        style_order=style_order,
        alpha=alpha,
        legend=legend,
        ax=ax,
        zorder=data_zorder,
        **scatter_kws,
    )
    ax = _deal_w_default_labels(ax, False)
    _circumplex_grid(ax, False, diagonal_lines, xlim, ylim)
    _set_circum_title(ax, False, title)
    if legend and hue:
        _move_legend(ax, legend_loc)
    return s


# TODO: Consider changing to displot
def density(data: Union[pd.DataFrame, np.ndarray] = None, x: str = "ISOPleasant", y: str = "ISOEventful",
            incl_scatter: bool = True, density_type: str = "full", title: Union[str, None] = "Soundscapy Density Plot",
            diagonal_lines: bool = False, xlim: tuple = (-1, 1), ylim: tuple = (-1, 1),
            scatter_kws: dict = dict(s=25, linewidth=0), incl_outline: bool = False, figsize: tuple = (5, 5),
            legend_loc: str = "lower left", alpha: float = 0.75, gridsize: int = 200, kernel: str = None,
            cut: Union[float, int] = 3, clip: Tuple[int] = None, legend: bool = False, cumulative: bool = False,
            cbar: bool = False, cbar_ax: matplotlib.axes.Axes = None, cbar_kws: dict = None,
            ax: matplotlib.axes.Axes = None, weights: str = None, hue: str = None, palette="colorblind",
            hue_order: List[str] = None, hue_norm=None, multiple: str = "layer", common_norm: bool = False,
            common_grid: bool = False, levels: int = 10, thresh: float = 0.05, bw_method="scott",
            bw_adjust: Union[float, int] = None, log_scale: Union[bool, int, float] = None, color: str = "blue",
            fill: bool = True, warn_singular: bool = True, **kwargs):
    """Plot a density plot of ISOCoordinates.

    Creates a wrapper around `seaborn.kdeplot` and adds functionality and styling to customise it for circumplex plots.
    The density plot is a combination of a kernel density estimate and a scatter plot.

    Parameters
    ----------
    data : pd.DataFrame, np.ndarray, mapping or sequence
        Input data structure. Either a long-form collection of vectors that can be assigned to
        named variables or a wide-form dataset that will be internally reshaped.
    x : vector or key in `data`, optional
        Column name for x variable, by default "ISOPleasant"
    y : vector or key in `data`, optional
        Column name for y variable, by default "ISOEventful"
    incl_scatter : bool, optional
        Whether to include a scatter plot of the data, by default True
    density_type : {"full", "simple"}, optional
        Type of density plot to draw, by default "full"
    title : str, optional
        Title to add to circumplex plot, by default "Soundscapy Density Plot"
    diagonal_lines : bool, optional
        Whether to include diagonal dimension labels (e.g. calm, etc.), by default False
    xlim, ylim : tuple, optional
        Limits of the circumplex plot, by default (-1, 1)
        It's recommended to set these such that the x and y axes have the same aspect
    scatter_kws : dict, optional
        Keyword arguments to pass to `seaborn.scatterplot`, by default dict(s=25, linewidth=0)
    incl_outline : bool, optional
    figsize : tuple, optional
        Size of the figure to return if `ax` is None, by default (5, 5)
    legend_loc : str, optional
        Relative location of legend, by default "lower left"
    alpha : float, optional
        Proportional opacity of the heatmap fill, by default 0.75
    gridsize : int, optional
        Nuber of points on each dimension of the evaluation grid, by default 200
    kernel : str, optional
        Function that defines the kernel, by default None
    cut : Union[float, int], optional
        Factor, multiplied by the smoothing bandwidth, that determines how far the evaluation grid extends past the
        extreme datapoints. When set to 0, truncate the curve at the data limits, by default 3
    clip : tuple[int], optional
        Do not evaluate the density outside of these limits, by default None
    legend : bool, optional
        If False, suppress the legend for semantic variables, by default True
    cumulative : bool, optional
        If True, estimate a cumulative distribution function, by default False
    cbar : bool, optional
        If True, add a colorbar to annotate the color mapping in a bivariate plot. Note: does not currently support
        plots with a `hue` variable well, by default False
    cbar_ax : matplotlib.axes.Axes, optional
        Pre-existing axes for the colorbar, by default None
    cbar_kws : dict, optional
        Keyword arguments for the colorbar, by default None
    ax : matplotlib.axes.Axes, optional
        Pre-existing axes object to use for the plot, by default None
    weights : vector or key in `data`, optional
        If provided, weight the kernel density estimation using these values, by default None
    hue : vector or key in `data`, optional
        Semantic variable that is mapped to determine the color of plot elements, by default None
    palette : Union[str, list, dict, matplotlib.colors.Colormap], optional
        Method for choosing the colors to use when mapping the hue semantic. String values are passed to
        seaborn.color_palette(). List or dict values imply categorical mapping, while a colormap object
        implies numeric mapping.
        by default colorblind
    hue_order : list[str], optional
        Specify the order of processing and plotting for categorical levels of the `hue` semantic, by default None
    hue_norm : Union[tuple, matplotlib.colors.Normalize], optional
        Either a pair of values that set the normalization range in data units or an object , by default None
    multiple : {"layer", "stack", "fill"}, optional
        Whether to plot multiple elements when semantic mapping creates subsets. Only relevant with univariate data,
        by default 'layer'
    common_norm : bool, optional
        If True, scale each conditional density by the number of observations such that the total area under all
        densities sums to 1. Otherwise, normalize each density independently, by default False
    common_grid : bool, optional
        If True, use the same evaluation grid for each kernel density estimate. Only relevant with univariate data.
        by default, False
    levels : int or vector, optional
        Number of contour levels or values to draw contours at. A vector argument must have increasing values in [0, 1].
         Levels correspond to iso-proportions of the density: e.g., 20% of the probability mass will lie below the
         contour drawn for 0.2. Only relevant with bivariate data.
         by default, 10
    thresh : number in [0, 1]
        Lowest iso-proportion level at which to draw a contour line. Ignored with `levels` is a vector. Only relevant
        with bivariate data.
    bw_method : string, scalar, or callable, optional
        Method for determining the smoothing bandwidth to use; passed to `scipy.stats.gaussian_kde`.
    bw_adjust : number, optional
        Factor that multiplicatively scales the value chosen using `bw_method`. Increasing will make the curve smoother.
        See Notes.
    log_scale : bool or number, or pair of bools or numbers, optional
        Set axis scale(s) to log. A single value sets the data axis for univariate distributions and both axes for
        bivariate distributions. A pair of values sets each axis independently. Numeric values are interpreted as the
        desired base (default 10). If False, defer to the existing Axes scale.
        by default None
    color : matplotlib color
        Single color specification for when hue mapping is not used. Otherwise the plot will try to hook into the
        matplotlib property cycle, by default "blue"
    fill : bool, optional
        If True, fill in the area under univariate density curves or between bivariate contours. If None, the default
        depends on `multiple`. by default True.
    warn_singular : bool, optional
        If True, issue a warning when trying to estimate the density of data with zero variance, by default True
    **kwargs : dict, optional#
        Other keyword arguments are passed to one of the following matplotlib functions:
        - `matplotlib.axes.Axes.plot()` (univariate, `fill=False`),
        - `matplotlib.axes.fill_between()` (univariate, `fill=True`),
        - `matplotlib.axes.Axes.contour()` (bivariate, `fill=True`),
        - `matplotlib.axes.Axes.contourf()` (bivariate, `fill=True`).

    Returns
    -------
    matplotlib.axes.Axes
        Axes object containing the plot.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)

    if bw_adjust is None:
        bw_adjust = default_bw_adjust

    if density_type == "simple":
        thresh = simple_density["thresh"]
        levels = simple_density["levels"]
        alpha = simple_density["alpha"]
        incl_outline = simple_density["incl_outline"]

    if incl_scatter:
        d = sns.scatterplot(
            data=data,
            x=x,
            y=y,
            hue=hue,
            ax=ax,
            palette=palette,
            zorder=data_zorder,
            **scatter_kws,
        )

    if incl_outline:
        d = sns.kdeplot(
            data=data,
            x=x,
            y=y,
            alpha=1,
            gridsize=gridsize,
            kernel=kernel,
            cut=cut,
            clip=clip,
            cumulative=cumulative,
            cbar=cbar,
            cbar_ax=cbar_ax,
            cbar_kws=cbar_kws,
            ax=ax,
            weights=weights,
            hue=hue,
            palette=palette,
            hue_order=hue_order,
            hue_norm=hue_norm,
            multiple=multiple,
            common_norm=common_norm,
            common_grid=common_grid,
            levels=levels,
            thresh=thresh,
            bw_method=bw_method,
            bw_adjust=bw_adjust,
            log_scale=log_scale,
            color=color,
            fill=False,
            warn_singular=warn_singular,
            zorder=data_zorder,
            **kwargs,
        )

    d = sns.kdeplot(
        data=data,
        x=x,
        y=y,
        alpha=alpha,
        gridsize=gridsize,
        kernel=kernel,
        cut=cut,
        clip=clip,
        legend=legend,
        cumulative=cumulative,
        cbar=cbar,
        cbar_ax=cbar_ax,
        cbar_kws=cbar_kws,
        ax=ax,
        weights=weights,
        hue=hue,
        palette=palette,
        hue_order=hue_order,
        hue_norm=hue_norm,
        multiple=multiple,
        common_norm=common_norm,
        common_grid=common_grid,
        levels=levels,
        thresh=thresh,
        bw_method=bw_method,
        bw_adjust=bw_adjust,
        log_scale=log_scale,
        color=color,
        fill=fill,
        warn_singular=warn_singular,
        zorder=data_zorder,
        **kwargs,
    )

    _circumplex_grid(
        ax, prim_labels=False, diagonal_lines=diagonal_lines, xlim=xlim, ylim=ylim
    )
    _set_circum_title(ax, prim_labels=False, title=title)
    _deal_w_default_labels(ax, prim_labels=False)
    if legend:
        _move_legend(ax, legend_loc)

    return d


def jointplot(
    data=None,
    x="ISOPleasant",
    y="ISOEventful",
    incl_scatter=True,
    density_type="full",
    title="Soundscape Joint Plot",
    diagonal_lines=False,
    xlim=(-1, 1),
    ylim=(-1, 1),
    scatter_kws=dict(s=25, linewidth=0),
    incl_outline=False,
    legend_loc="lower left",
    alpha=0.75,
    color=None,
    joint_kws={},
    marginal_kws={"fill": True, "common_norm": False},
    hue=None,
    palette="colorblind",
    hue_order=None,
    hue_norm=None,
    common_norm=False,
    fill=True,
    bw_adjust=None,
    thresh=0.1,
    levels=10,
    legend=False,
    marginal_kind="kde",
):
    """Create a jointplot with distribution or scatter in the center and distributions on the margins.

    This method works by calling sns.jointplot() and creating a circumplex grid in the joint position, then
    overlaying a density or circumplex_scatter plot. The options for both the joint and marginal plots can be
    passed through the sns.jointplot() separately to customise them separately. The marginal distribution plots
    can be either a density or histogram.

    Parameters
    ----------
    data : pd.DataFrame, np.ndarray, mapping, or sequence
        Input data structure. Either a long-form collection of vectors that can be assigned to named variables or a
        wide-form dataset that will be internally reshaped.
    x : vector or key in `data`, optional
        column name for x variable, by default "ISOPleasant"
    y : vector or key in `data`, optional
        column name for y variable, by default "ISOEventful"
    incl_scatter : bool, optional
        Whether to include a scatter plot of the data, by default True
    density_type : str, optional
        Type of density plot to draw, by default "full"
    diagonal_lines : bool, optional
        whether to include diagonal dimension axis labels in the joint plot, by default False
    palette : str, optional
        [description], by default "colorblind"
    incl_scatter : bool, optional
        plot coordinate scatter underneath density plot, by default False
    fill : bool, optional
        whether to fill the density plot, by default True
    bw_adjust : [type], optional
        [description], by default default_bw_adjust
    alpha : float, optional
        [description], by default 0.95
    legend : bool, optional
        whether to include the hue labels legend, by default False
    legend_loc : str, optional
        relative location of the legend, by default "lower left"
    marginal_kind : str, optional
        density or histogram plot in the margins, by default "kde"
    hue : vector or key in data, optional
        Grouping variable that will produce points with different colors. Can be either categorical or numeric,
        although color mapping will behave differently in latter case, by default None
    joint_kws : dict, optional
        Arguments to pass to density or scatter joint plot, by default {}
    marginal_kws : dict, optional
        Arguments to pass to marginal distribution plots, by default {"fill": True}
    hue : vector or key in `data`, optional
        Semantic variable that is mapped to determine the color of plot elements.
    palette : string, list, dict, or `matplotlib.colors.Colormap`, optional
        Method for choosing the colors to use when mapping the `hue` semantic. String values are passed to
        `color_palette()`. List or dict values imply categorical mapping, while a colormap object implies numeric
        mapping.
        by default, `"colorblind"`
    hue_order : vector of strings, optional.
        Specify the order of processing and plotting for categorical levels of the `hue` semantic.
    hue_norm : tuple or matplotlib.colors.Normalize, optional
        Either a pair of values that set the normalization range in data units or an object that will map from data
        units into a [0, 1] interval. Usage implies numeric mapping.
    common_norm : bool, optional
        If True, scale each conditional density by the number of observations such that the total area under all
        densities sums to 1. Otherwise, normalize each density independently, by default False.
    fill : bool, optional
        If True, fill in the area under univariate density curves or between bivariate contours. If None, the default
        depends on `multiple`. by default True
    bw_adjust : number, optional
        Factor that multiplicatively scales the value chosen using `bw_method`. Increasing will make the curve smoother.
        See Notes. by default default_bw_adjust (1.2)
    thresh : number in [0, 1], optional
        Lowest iso-proportional level at which to draw a contour line. Ignored when `levels` is a vector. Only relevant
        with bivariate plots. by default 0.1
    levels : int or vector, optional
        Number of contour levels or values to draw contours at. A vector argument must have increasing values in [0, 1].
        Levels correspond to iso-proportionas of the density: e.g. 20% of the probability mass will lie below the
        contour drawn for 0.2. Only relevant with bivariate data.
        by default 10
    legend : bool, optional
        If False, suppress the legend for semantic variables, by default False
    legend_loc : str, optional
        Relative location of the legend, by default "lower left"
    marginal_kind : str, optional
        density or histogram plot in the margins, by default "kde"

    Returns
    -------
    plt.Axes
    """

    if bw_adjust is None:
        bw_adjust = default_bw_adjust

    if density_type == "simple":
        thresh = simple_density["thresh"]
        levels = simple_density["levels"]
        alpha = simple_density["alpha"]
        incl_outline = simple_density["incl_outline"]

    g = sns.JointGrid()
    density(data, x=x, y=y, incl_scatter=incl_scatter, density_type=density_type, title=None,
            diagonal_lines=diagonal_lines, xlim=xlim, ylim=ylim, scatter_kws=scatter_kws, incl_outline=incl_outline,
            legend_loc=legend_loc, alpha=alpha, legend=legend, ax=g.ax_joint, hue=hue, palette=palette,
            hue_order=hue_order, hue_norm=hue_norm, common_norm=common_norm, levels=levels, thresh=thresh,
            bw_adjust=bw_adjust, color=color, fill=fill, **joint_kws)
    # if legend and hue:
    #     _move_legend(g.ax_joint, legend_loc)

    if marginal_kind == "hist":
        sns.histplot(
            data=data,
            x=x,
            hue=hue,
            palette=palette,
            ax=g.ax_marg_x,
            binrange=xlim,
            legend=False,
            **marginal_kws,
        )
        sns.histplot(
            data=data,
            y=y,
            hue=hue,
            palette=palette,
            ax=g.ax_marg_y,
            binrange=ylim,
            legend=False,
            **marginal_kws,
        )
    elif marginal_kind == "kde":
        sns.kdeplot(
            data=data,
            x=x,
            hue=hue,
            palette=palette,
            ax=g.ax_marg_x,
            bw_adjust=bw_adjust,
            legend=False,
            **marginal_kws,
        )
        sns.kdeplot(
            data=data,
            y=y,
            hue=hue,
            palette=palette,
            ax=g.ax_marg_y,
            bw_adjust=bw_adjust,
            legend=False,
            **marginal_kws,
        )
    g.ax_marg_x.set_title(title, pad=6.0)

    return g


def _move_legend(ax, new_loc, **kws):
    """Moves legend to desired relative location.

    Parameters
    ----------
    ax : plt.Axes
        Existing axes object to adjust the legend on
    new_loc : str or pair of floats
        The location of the legend
    """
    old_legend = ax.legend_
    handles = old_legend.legendHandles
    labels = [t.get_text() for t in old_legend.get_texts()]
    title = old_legend.get_title().get_text()
    ax.legend(handles, labels, loc=new_loc, title=title, **kws)


def _circumplex_grid(
    ax, prim_labels=True, diagonal_lines=False, xlim=(-1, 1), ylim=(-1, 1)
):
    """Create the base layer grids and label lines for the soundscape circumplex

    Parameters
    ----------
    axes : matplotlib.pyplot.Axes
        plt subplot Axes to add the circumplex grids to
    prim_labsl: bool, optional
        flag for whether to include the custom primary labels ISOPleasant and ISOEventful, by default True
        If using your own x and y names, you should set this to False.
    diagonal_lines : bool, optional
        flag for whether the include the diagonal dimensions (calm, etc), by default False
    """
    # Setting up the grids
    sns.set_style({"xtick.direction": "in", "ytick.direction": "in"})
    line_weights = 1.5
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    # grids and ticks
    ax.get_xaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())
    ax.get_yaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())

    ax.grid(visible=True, which="major", color="grey", alpha=0.5)
    ax.grid(
        visible=True,
        which="minor",
        color="grey",
        linestyle="dashed",
        linewidth=0.5,
        alpha=0.4,
        zorder=prim_lines_zorder,
    )

    ax = _primary_lines_and_labels(ax, prim_labels, line_weights=line_weights)
    if diagonal_lines:
        ax = _diagonal_lines_and_labels(ax, line_weights=line_weights)

    return ax


def _set_circum_title(ax, prim_labels, title):
    """Set the title for the circumplex plot

    Parameters
    ----------
    ax : plt.Axes
        Existing axes object to adjust the legend on
    prim_labels: bool, optional
        flag for whether to include the custom primary labels ISOPleasant and ISOEventful, by default True
        If using your own x and y names, you should set this to False.
    title : str
        Title to set
    """
    title_pad = 30.0 if prim_labels is True else 6.0
    ax.set_title(title, pad=title_pad)
    return ax


def _deal_w_default_labels(ax, prim_labels):
    """Deal with the default labels for the circumplex plot

    Parameters
    ----------
    ax : plt.Axes
        Existing axes object to adjust the legend on
    prim_labels: bool, optional
        flag for whether to include the custom primary labels ISOPleasant and ISOEventful, by default True
        If using your own x and y names, you should set this to False.
    """
    if prim_labels is True or prim_labels == "none":
        # hide axis labels
        ax.xaxis.label.set_visible(False)
        ax.yaxis.label.set_visible(False)

    return ax


def _primary_lines_and_labels(ax, prim_labels, line_weights):
    x_lim = ax.get_xlim()
    y_lim = ax.get_ylim()

    # Add lines and labels for circumplex model
    ## Primary Axes
    ax.plot(  # Horizontal line
        [x_lim[0], x_lim[1]],
        [0, 0],
        linestyle="dashed",
        color="grey",
        alpha=1,
        lw=line_weights,
        zorder=prim_lines_zorder,
    )
    ax.plot(  # vertical line
        [0, 0],
        [y_lim[0], y_lim[1]],
        linestyle="dashed",
        color="grey",
        alpha=1,
        lw=line_weights,
        zorder=prim_lines_zorder,
    )

    if prim_labels is True:
        prim_ax_font = {
            "fontstyle": "italic",
            "fontsize": "medium",
            "fontweight": "bold",
            "c": "grey",
            "alpha": 1,
        }
        ### Labels
        ax.text(  # ISOPleasant Label
            x=x_lim[1] + 0.01,
            y=0,
            s="ISO\nPleasant",
            ha="left",
            va="center",
            fontdict=prim_ax_font,
        )
        ax.text(  # ISOEventful Label
            x=0,
            y=y_lim[1] + 0.01,
            s="ISO\nEventful",
            ha="center",
            va="bottom",
            fontdict=prim_ax_font,
        )

    return ax


def _diagonal_lines_and_labels(ax, line_weights):
    diag_ax_font = {
        "fontstyle": "italic",
        "fontsize": "small",
        "fontweight": "bold",
        "c": "black",
        "alpha": 0.5,
    }
    x_lim = ax.get_xlim()
    y_lim = ax.get_ylim()

    ax.plot(  # uppward diagonal
        [x_lim[0], x_lim[1]],
        [y_lim[0], y_lim[1]],
        linestyle="dashed",
        color="grey",
        alpha=0.5,
        lw=line_weights,
        zorder=diag_lines_zorder,
    )
    ax.plot(  # downward diagonal
        [x_lim[0], x_lim[1]],
        [y_lim[1], y_lim[0]],
        linestyle="dashed",
        color="grey",
        alpha=0.5,
        lw=line_weights,
        zorder=diag_lines_zorder,
    )

    ### Labels
    ax.text(  # Vibrant label
        x=x_lim[1] / 2,
        y=y_lim[1] / 2,
        s="(vibrant)",
        ha="center",
        va="center",
        fontdict=diag_ax_font,
        zorder=diag_labels_zorder,
    )
    ax.text(  # Chaotic label
        x=x_lim[0] / 2,
        y=y_lim[1] / 2,
        s="(chaotic)",
        ha="center",
        va="center",
        fontdict=diag_ax_font,
        zorder=diag_labels_zorder,
    )
    ax.text(  # monotonous label
        x=x_lim[0] / 2,
        y=y_lim[0] / 2,
        s="(monotonous)",
        ha="center",
        va="center",
        fontdict=diag_ax_font,
        zorder=diag_labels_zorder,
    )
    ax.text(  # calm label
        x=x_lim[1] / 2,
        y=y_lim[0] / 2,
        s="(calm)",
        ha="center",
        va="center",
        fontdict=diag_ax_font,
        zorder=diag_labels_zorder,
    )
    return ax


# Unsupported right now.


def single_lmplot(
    sf,
    x,
    y,
    groups=None,
    group_order=None,
    ylim=None,
    xlim=None,
    order=1,
    fit_reg=True,
    corr="pearson",
    **kwargs,
):
    """Unfinished and unsupported"""
    if y in ["ISOPleasant", "ISOEventful"] and ylim is None:
        ylim = (-1, 1)
    if corr:
        df = sf[[y, x]].dropna()
        if corr == "pearson":
            r, p = pearsonr(df[y], df[x])
        elif corr == "spearman":
            r, p == spearmanr(df[y], df[x])

        if p < 0.01:
            symb = "**"
        elif p < 0.05:
            symb = "*"
        else:
            symb = ""
        text = f"{round(r, 2)}{symb}"
    else:
        text = None

    if groups is None:
        g = sns.lmplot(
            x=x,
            y=y,
            data=sf,
            height=8,
            aspect=1,
            order=order,
            fit_reg=fit_reg,
            **kwargs,
        )
    else:
        g = sns.lmplot(
            x=x,
            y=y,
            data=sf,
            height=8,
            aspect=1,
            order=order,
            hue=groups,
            hue_order=group_order,
            fit_reg=fit_reg,
            **kwargs,
        )

    g.set(ylim=ylim)
    g.set(xlim=xlim)

    ax = g.axes.ravel()[0]
    ax.text(
        np.mean(ax.get_xlim()),
        min(ax.get_ylim()) * 0.75,
        text,
        ha="center",
        fontweight=750,
    )

    return g


def grouped_lmplot(
    sf,
    x,
    y,
    groups,
    group_order=None,
    ylim=None,
    xlim=None,
    order=1,
    fit_reg=True,
    corr="pearson",
    col_wrap=4,
    scatter_kws=None,
    line_kws=None,
    facet_kws={"sharex": False},
    **kwargs,
):
    """Unfinished and unsupported"""
    if y in ["ISOPleasant", "ISOEventful"] and ylim is None:
        ylim = (-1, 1)
    grid = sns.lmplot(
        x=x,
        y=y,
        col=groups,
        order=order,
        col_wrap=col_wrap,
        data=sf,
        fit_reg=fit_reg,
        height=4,
        facet_kws=facet_kws,
        scatter_kws=scatter_kws,
        line_kws=line_kws,
        **kwargs,
    )
    grid.set(ylim=ylim)
    grid.set(xlim=xlim)
    if group_order is None:
        group_order = sf[groups].unique()

    if corr:
        for location, ax in zip(group_order, grid.axes.ravel()):
            df = sf.loc[sf[groups] == location, [y, x]].dropna()
            if corr == "pearson":
                r, p = pearsonr(df[y], df[x])
            elif corr == "spearman":
                r, p = spearmanr(df[y], df[x])
            if p < 0.01:
                symb = "**"
            elif p < 0.05:
                symb = "*"
            else:
                symb = ""
            fontweight = 750 if p < 0.05 else None

            text = f"{round(r, 2)}{symb}"
            ax.text(
                np.mean(ax.get_xlim()),
                min(ax.get_ylim()) * 0.75,
                text,
                ha="center",
                fontweight=fontweight,
            )
    grid.set_titles("{col_name}")
    return grid


def iso_annotation(
    ax,
    data,
    location,
    x_adj=0,
    y_adj=0,
    x_key="ISOPleasant",
    y_key="ISOEventful",
    ha="center",
    va="center",
    fontsize="small",
    arrowprops=dict(arrowstyle="-", ec="black"),
    **text_kwargs,
):
    """add text annotations to circumplex plot based on coordinate values

    Directly uses plt.annotate

    Parameters
    ----------
    ax : matplotlib.pyplot.Axes
        existing plt axes to add to
    data : pd.Dataframe
        dataframe of coordinate points
    location : str
        name of the coordinate to plot
    x_adj : int, optional
        value to adjust x location by, by default 0
    y_adj : int, optional
        value to adjust y location by, by default 0
    x_key : str, optional
        name of x column, by default "ISOPleasant"
    y_key : str, optional
        name of y column, by default "ISOEventful"
    ha : str, optional
        horizontal alignment, by default "center"
    va : str, optional
        vertical alignment, by default "center"
    fontsize : str, optional
        by default "small"
    arrowprops : dict, optional
        dict of properties to send to plt.annotate, by default dict(arrowstyle="-", ec="black")

    Example
    -------
    >>> fig, axes = plt.subplots(1,1, figsize=(5,5))
    >>> df_mean.isd.scatter(xlim=(-.5, .5),ylim=(-.5, .5),ax=axes)
    >>> for location in df_mean.LocationID:
    >>>     plotting.iso_annotation(axes, df_mean, location)
    """
    ax.annotate(
        text=data["LocationID"][location],
        xy=(
            data[x_key][location],
            data[y_key][location],
        ),
        xytext=(
            data[x_key][location] + x_adj,
            data[y_key][location] + y_adj,
        ),
        ha=ha,
        va=va,
        arrowprops=arrowprops,
        annotation_clip=True,
        fontsize=fontsize,
        **text_kwargs,
    )



# Constants and Labels
PAQ_NAMES = [
    "pleasant",
    "vibrant",
    "eventful",
    "chaotic",
    "annoying",
    "monotonous",
    "uneventful",
    "calm",
]

PAQ_IDS = [
    "PAQ1",
    "PAQ2",
    "PAQ3",
    "PAQ4",
    "PAQ5",
    "PAQ6",
    "PAQ7",
    "PAQ8",
]

IGNORE_LIST = ["AllLondon", "AllyPally", "CoventGd1", "OxfordSt"]

CATEGORISED_VARS = {
    "indexing": [
        "GroupID",
        "SessionID",
        "LocationID",
        "Country",
        "record_id",
    ],  # Ways to index which survey it is
    "meta_info": [
        "recording",
        "start_time",
        "end_time",
        "longitude",
        "latitude",
    ],  # Info from when that survey was collected
    "sound_source_dominance": [
        "Traffic",
        "Other",
        "Human",
        "Natural",
    ],  # Sound sources
    "complex_PAQs": ["Pleasant", "Eventful"],  # Projected PAQ coordinates
    "raw_PAQs": [
        "pleasant",
        "chaotic",
        "vibrant",
        "uneventful",
        "calm",
        "annoying",
        "eventful",
        "monotonous",
    ],  # Raw 5-point PAQs
    "overall_soundscape": [
        "overall",
        "appropriateness",
        "loudness",
        "often",
        "visit_again",
    ],  # Questions about the overall soundscape
    "demographics": ["Age", "Gender", "Occupation", "Education", "Ethnicity", "Resid"],
    "misc": ["AnythingElse"],
}



"""Soundscapy Dataframe Accessor

This module contains the Soundscapy Dataframe Accessor, which is used to add methods to Pandas Dataframes.
This method is used to simplify the direct manipulation of soundscape survey-type data, allowing methods to be chained.
The accessor is added to the dataframe by importing the module, and is accessed by using the `sspy` attribute of the dataframe.
The sspy accessor is intended to be general to any soundscape survey data, functions for specific survey types are contained in the `isd` and `araus` modules.
"""

# Add soundscapy to the Python path
import sys
from datetime import date
from typing import Union, Tuple, List, Dict

import matplotlib
import numpy as np
import pandas as pd
from pandas.api.extensions import register_dataframe_accessor

import soundscapy.plotting.likert

import soundscapy.utils.surveys as db
import soundscapy.plotting.circumplex as sspyplot

# Define the names of the PAQ columns


# Default plot settings
diag_lines_zorder = 1
diag_labels_zorder = 4
prim_lines_zorder = 2
data_zorder = 3
default_bw_adjust = 1.2


@register_dataframe_accessor("sspy")
class SSPYAccessor:
    """Soundscapy Dataframe Accessor

    This class is used to add methods to Pandas Dataframes.
    This method is used to simplify the direct manipulation of soundscape survey-type data, allowing methods to be chained.
    The accessor is added to the dataframe by importing the module, and is accessed by using the `sspy` attribute of the dataframe.
    The sspy accessor is intended to be general to any soundscape survey data, functions for specific survey types are
    contained in the `isd` and `araus` modules.
    """

    def __init__(self, df):
        self._df = df
        self._analysis_date = date.today().isoformat()
        self._metadata = {}

    def validate_dataset(
        self,
        paq_aliases: Union[List, Dict] = None,
        allow_lockdown: bool = True,
        allow_na: bool = False,
        verbose: int = 1,
        val_range: Tuple = (1, 5),
    ):
        """Performs data quality checks and validates that the dataset fits the expected format

        Parameters
        ----------
        df : pd.DataFrame
            ISD style dataframe, incl PAQ data
        paq_aliases : list or dict, optional
            list of PAQ names (in order)
            or dict of PAQ names with new names as values, by default None
        allow_lockdown : bool, optional
            if True will keep Lockdown data in the df, by default True
        allow_paq_na : bool, optional
            remove rows which have any missing PAQ values
            otherwise will remove those with 50% missing, by default False    verbose : int, optional
            how much info to print while running, by default 1
        val_range : tuple, optional
            min and max range of the PAQ response values, by default (5, 1)

        Returns
        -------
        tuple
            cleaned dataframe, dataframe of excluded samples

        See Also
        --------
        :func:`soundscapy.database.validate_dataset`
        """
        return db.validate_dataset(
            self._df, paq_aliases, allow_lockdown, allow_na, verbose, val_range
        )

    def paq_data_quality(self, verbose=0):
        """
        Check the data quality of the PAQs in the dataframe

        Parameters
        ----------
        verbose : int, optional
            How much info to print, by default 0

        Returns
        -------
        pd.DataFrame
            Dataframe of the data quality checks

        See Also
        --------
        :func:`soundscapy.database.paq_data_quality`
        """
        return db.paq_data_quality(self._df, verbose)

    def filter(self, filter_by, condition):
        """Filter the dataframe by a condition

        Parameters
        ----------
        filter_by : str
            Column to filter by
        condition : str
            Condition to filter by

        Returns
        -------
        pd.DataFrame
        """
        if isinstance(condition, str):
            return self._df.query(f"{filter_by} == @condition")
        elif isinstance(condition, (list, tuple)):
            return self._df.query(f"{filter_by} in @condition")

    # TODO: add mean_responses method

    def convert_column_to_index(self, col, drop=False):
        """Convert a column to the index of the dataframe

        Parameters
        ----------
        col : str
            Column to convert to index
        drop : bool, optional
            Drop the column after conversion, by default False

        Returns
        -------
        pd.DataFrame
            Dataframe with the column converted to the index

        See Also
        --------
        :func:`soundscapy.database.convert_column_to_index`
        """
        return db.convert_column_to_index(self._df, col, drop)

    def return_paqs(self, incl_ids=True, other_cols=None):
        """Return only the PAQ columns

        Parameters
        ----------
        incl_ids : bool, optional
            whether to include ID cols too (i.e. RecordID, GroupID, etc), by default True
        other_cols : list, optional
            other columns to also include, by default None

        Returns
        -------
        pd.DataFrame
            Dataframe with only the PAQ columns

        See Also
        --------
        :func:`soundscapy.database.return_paqs`
        """
        return db.return_paqs(self._df, incl_ids, other_cols)

    def add_paq_coords(
        self,
        scale_to_one: bool = True,
        val_range=(5, 1),
        projection: bool = True,
        names=("ISOPleasant", "ISOEventful"),
    ):
        """Calculate and add ISO coordinates as new columns in dataframe

        Calls `calculate_paq_coords()`

        Parameters
        ----------
        scale_to_one : bool, optional
            Should the coordinates be scaled to (-1, +1), by default True
        val_range: tuple, optional
            (max, min) range of original PAQ responses, by default (5, 1)
        projection : bool, optional
            Use the trigonometric projection (cos(45)) term for diagonal PAQs, by default True
        names : list, optional
            Names for new coordinate columns, by default ["ISOPleasant", "ISOEventful"]

        Returns
        -------
        pd.DataFrame
            Dataframe with new columns added

        See Also
        --------
        :func:`soundscapy.database.calculate_paq_coords`
        """
        isopl, isoev = db.calculate_paq_coords(
            self._df, scale_to_one, val_range, projection
        )
        return self._df.add_column(names[0], isopl).add_column(names[1], isoev)

    def soundscapy_describe(self, group_by, type="percent"):
        """Describe the dataframe by a column

        Parameters
        ----------
        group_by : str
            Column to group by
        type : str, optional
            Type of summary to return, by default "percent"

        Returns
        -------
        pd.DataFrame
            Dataframe of summary stats

        See Also
        --------
        :func:`soundscapy.database.soundscapy_describe`
        """
        res = {
            location: self.location_describe(location, type=type)
            for location in self._df[group_by].unique()
        }

        res = pd.DataFrame.from_dict(res, orient="index")
        return res

    def paq_radar(self, ax=None, index=None):
        """Plot a radar chart of the PAQs

        Parameters
        ----------
        ax : matplotlib.axes, optional
            Axis to plot on, by default None
        index : int, optional
            Index of the row to plot, by default None

        Returns
        -------
        matplotlib.axes
            Axis with the plot

        See Also
        --------
        :func:`soundscapy.database.paq_radar`
        """
        return soundscapy.plotting.likert.paq_radar_plot(self._df, ax, index)

    def scatter(
        self,
        x="ISOPleasant",
        y="ISOEventful",
        title="Soundscape Scatter Plot",
        diagonal_lines=False,
        xlim=(-1, 1),
        ylim=(-1, 1),
        figsize=(5, 5),
        legend_loc="lower left",
        hue=None,
        style=None,
        s=10,
        palette="colorblind",
        hue_order=None,
        hue_norm=None,
        sizes=None,
        size_order=None,
        size_norm=None,
        markers=True,
        style_order=None,
        alpha=None,
        legend="auto",
        ax=None,
        **scatter_kwargs,
    ):
        """Plot ISOcoordinates as scatter points on a soundscape circumplex grid

        Makes use of seaborn.scatterplot


         - Soundscapy specific parameters -
        We have made all of the `seaborn.scatterplot` arguments available, but have also added or changed some specific
        options for circumplex plotting.


        Parameters
        ----------
        x : vector or key in `data`, optional
            column name for x variable, by default "ISOPleasant"
        y : vector or key in `data`, optional
            column name for y variable, by default "ISOEventful"
        title : str, optional
            Title to add to circumplex plot, by default "Soundscape Scatter Plot"
        diagonal_lines : bool, optional
            whether to include diagonal dimension labels (e.g. calm, etc.), by default False
        xlim, ylim : tuple, optional
            Limits of the circumplex plot, by default (-1, 1)
            It's recommended to set these such that the x and y axes have the same aspect
        figsize : tuple, optional
            Size of the figure to return if `ax` is None, by default (5, 5)
        legend_loc : str, optional
            relative location of legend, by default "lower left"
        palette : string, list, dict or matplotlib.colors.Colormap, optional
            Method for choosing the colors to use when mapping the hue semantic. String values are passed to
            seaborn.color_palette(). List or dict values imply categorical mapping, while a colormap object
            implies numeric mapping.
            by default colorblind
        s : int, optional
            size of scatter points, by default 10
        hue : vector or key in data, optional
            Grouping variable that will produce points with different colors. Can be either categorical or numeric,
            although color mapping will behave differently in latter case, by default None
        style : vector or key in data
            Grouping variable that will produce points with different markers. Can have a numeric dtype but will always
            be treated as categorical.
        hue_order : vector of strings
            Specify the order of processing and plotting for categorical levels of the `hue` semantic
        hue_norm : tuple or matplotlib.colors.Normalize
            Either a pair of values that set the normalization range in data units or an object that will map from data
            units into a [0, 1] interval. Usage implies numeric mapping.
        sizes : list, dict, or tuple
            An object that determines how sizes are chosen when `size` is used. It can always be a list of size values or
            a dict mapping levels of the `size` variable to sizes. When `size` is numeric, it can also be a tuple
            specifying the minimum and maximum size to use such that other values are normalized within this range.
        size_order : list
            Specified order for appearance of the `size` variable levels, otherwise they are determined from the data. Not
            relevant when the `size` variable is numeric.
        size_norm : tuple or Normalization object
            Normalization in data units for scaling plot objects when the `size` variable is numeric.
        markers : boolean, list, or dictionary
            Object determining how to draw the markers for different levels of the `style` variable. Setting to `True` will
            use default markers, or you can pass a list of markers or a dictionary mapping levels of the `style` variable
            to markers. Setting to `False` will draw marker-less lines. Markers are specified as in matplotlib.
        style_order : list
            Specified order for appearance of the `style` variable levels otherwise they are determined from the data. Not
            relevant when the `style` variable is numeric.
        alpha : float
            Proportional opacity of the points.
        legend : {"auto", "brief", "full" or False}, optional
            How to draw the legend. If “brief”, numeric hue and size variables will be represented with a sample of evenly
            spaced values. If “full”, every group will get an entry in the legend. If “auto”, choose between brief or full
            representation based on number of levels. If False, no legend data is added and no legend is drawn.
            by default, "auto"
        ax : matplotlib.axes.Axes, optional
            Pre-existing matplotlib axes for the plot, by default None
            If `None` call `matplotlib.pyplot.subplots` with `figsize` internally.

        Returns
        -------
        matplotlib.axes.Axes

        See Also
        --------
        :func:`soundscapy.plotting.circumplex.scatter`

        `seaborn.kdeplot <https://seaborn.pydata.org/generated/seaborn.kdeplot.html>`_
        """
        return sspyplot.scatter(
            self._df,
            x=x,
            y=y,
            title=title,
            diagonal_lines=diagonal_lines,
            xlim=xlim,
            ylim=ylim,
            figsize=figsize,
            legend_loc=legend_loc,
            hue=hue,
            style=style,
            s=s,
            palette=palette,
            hue_order=hue_order,
            hue_norm=hue_norm,
            sizes=sizes,
            size_order=size_order,
            size_norm=size_norm,
            markers=markers,
            style_order=style_order,
            alpha=alpha,
            legend=legend,
            ax=ax,
            **scatter_kwargs,
        )

    def density(
        self,
        x: str = "ISOPleasant",
        y: str = "ISOEventful",
        incl_scatter: bool = True,
        density_type: str = "full",
        title: str = "Soundscapy Density Plot",
        diagonal_lines: bool = False,
        xlim: tuple = (-1, 1),
        ylim: tuple = (-1, 1),
        scatter_kws: dict = dict(s=20),
        incl_outline: bool = False,
        figsize: tuple = (5, 5),
        legend_loc: str = "lower left",
        alpha: float = 0.75,
        gridsize: int = 200,
        kernel: str = None,
        cut: Union[float, int] = 3,
        clip: Tuple[int] = None,
        legend: bool = False,
        cumulative: bool = False,
        cbar: bool = False,
        cbar_ax: matplotlib.axes.Axes = None,
        cbar_kws: dict = None,
        ax: matplotlib.axes.Axes = None,
        weights: str = None,
        hue: str = None,
        palette="colorblind",
        hue_order: List[str] = None,
        hue_norm=None,
        multiple: str = "layer",
        common_norm: bool = False,
        common_grid: bool = False,
        levels: int = 10,
        thresh: float = 0.05,
        bw_method="scott",
        bw_adjust: Union[float, int] = None,
        log_scale: Union[bool, int, float] = None,
        color: str = "blue",
        fill: bool = True,
        data2: Union[pd.DataFrame, np.ndarray] = None,
        warn_singular: bool = True,
        **kwargs,
    ):  # sourcery skip: default-mutable-arg
        """Plot a density plot of ISOCoordinates.

        Creates a wrapper around `seaborn.kdeplot` and adds functionality and styling to customise it for circumplex plots.
        The density plot is a combination of a kernel density estimate and a scatter plot.

         - Soundscapy specific parameters -
        We have made all of the `seaborn.scatterplot` arguments available, but have also added or changed some specific
        options for circumplex plotting.

        Parameters
        ----------
        data : pd.DataFrame, np.ndarray, mapping or sequence
            Input data structure. Either a long-form collection of vectors that can be assigned to
            named variables or a wide-form dataset that will be internally reshaped.
        x : vector or key in `data`, optional
            Column name for x variable, by default "ISOPleasant"
        y : vector or key in `data`, optional
            Column name for y variable, by default "ISOEventful"
        incl_scatter : bool, optional
            Whether to include a scatter plot of the data, by default True
        density_type : {"full", "simple"}, optional
            Type of density plot to draw, by default "full"
        title : str, optional
            Title to add to circumplex plot, by default "Soundscapy Density Plot"
        diagonal_lines : bool, optional
            Whether to include diagonal dimension labels (e.g. calm, etc.), by default False
        xlim, ylim : tuple, optional
            Limits of the circumplex plot, by default (-1, 1)
            It's recommended to set these such that the x and y axes have the same aspect
        scatter_kws : dict, optional
            Keyword arguments to pass to `seaborn.scatterplot`, by default dict(s=25, linewidth=0)
        incl_outline : bool, optional
        figsize : tuple, optional
            Size of the figure to return if `ax` is None, by default (5, 5)
        legend_loc : str, optional
            Relative location of legend, by default "lower left"
        alpha : float, optional
            Proportional opacity of the heatmap fill, by default 0.75
        gridsize : int, optional
            Nuber of points on each dimension of the evaluation grid, by default 200
        kernel : str, optional
            Function that defines the kernel, by default None
        cut : Union[float, int], optional
            Factor, multiplied by the smoothing bandwidth, that determines how far the evaluation grid extends past the
            extreme datapoints. When set to 0, truncate the curve at the data limits, by default 3
        clip : tuple[int], optional
            Do not evaluate the density outside of these limits, by default None
        legend : bool, optional
            If False, suppress the legend for semantic variables, by default True
        cumulative : bool, optional
            If True, estimate a cumulative distribution function, by default False
        cbar : bool, optional
            If True, add a colorbar to annotate the color mapping in a bivariate plot. Note: does not currently support
            plots with a `hue` variable well, by default False
        cbar_ax : matplotlib.axes.Axes, optional
            Pre-existing axes for the colorbar, by default None
        cbar_kws : dict, optional
            Keyword arguments for the colorbar, by default None
        ax : matplotlib.axes.Axes, optional
            Pre-existing axes object to use for the plot, by default None
        weights : vector or key in `data`, optional
            If provided, weight the kernel density estimation using these values, by default None
        hue : vector or key in `data`, optional
            Semantic variable that is mapped to determine the color of plot elements, by default None
        palette : Union[str, list, dict, matplotlib.colors.Colormap], optional
            Method for choosing the colors to use when mapping the hue semantic. String values are passed to
            seaborn.color_palette(). List or dict values imply categorical mapping, while a colormap object
            implies numeric mapping.
            by default colorblind
        hue_order : list[str], optional
            Specify the order of processing and plotting for categorical levels of the `hue` semantic, by default None
        hue_norm : Union[tuple, matplotlib.colors.Normalize], optional
            Either a pair of values that set the normalization range in data units or an object , by default None
        multiple : {"layer", "stack", "fill"}, optional
            Whether to plot multiple elements when semantic mapping creates subsets. Only relevant with univariate data,
            by default 'layer'
        common_norm : bool, optional
            If True, scale each conditional density by the number of observations such that the total area under all
            densities sums to 1. Otherwise, normalize each density independently, by default False
        common_grid : bool, optional
            If True, use the same evaluation grid for each kernel density estimate. Only relevant with univariate data.
            by default, False
        levels : int or vector, optional
            Number of contour levels or values to draw contours at. A vector argument must have increasing values in [0, 1].
             Levels correspond to iso-proportions of the density: e.g., 20% of the probability mass will lie below the
             contour drawn for 0.2. Only relevant with bivariate data.
             by default, 10
        thresh : number in [0, 1]
            Lowest iso-proportion level at which to draw a contour line. Ignored with `levels` is a vector. Only relevant
            with bivariate data.
        bw_method : string, scalar, or callable, optional
            Method for determining the smoothing bandwidth to use; passed to `scipy.stats.gaussian_kde`.
        bw_adjust : number, optional
            Factor that multiplicatively scales the value chosen using `bw_method`. Increasing will make the curve smoother.
            See Notes.
        log_scale : bool or number, or pair of bools or numbers, optional
            Set axis scale(s) to log. A single value sets the data axis for univariate distributions and both axes for
            bivariate distributions. A pair of values sets each axis independently. Numeric values are interpreted as the
            desired base (default 10). If False, defer to the existing Axes scale.
            by default None
        color : matplotlib color
            Single color specification for when hue mapping is not used. Otherwise the plot will try to hook into the
            matplotlib property cycle, by default "blue"
        fill : bool, optional
            If True, fill in the area under univariate density curves or between bivariate contours. If None, the default
            depends on `multiple`. by default True.
        data2 : Union[pd.DataFrame, np.ndarray] optional
            Second set of data to plot when `multiple` is "stack" or "fill".
        warn_singular : bool, optional
            If True, issue a warning when trying to estimate the density of data with zero variance, by default True
        **kwargs : dict, optional#
            Other keyword arguments are passed to one of the following matplotlib functions:
            - `matplotlib.axes.Axes.plot()` (univariate, `fill=False`),
            - `matplotlib.axes.fill_between()` (univariate, `fill=True`),
            - `matplotlib.axes.Axes.contour()` (bivariate, `fill=True`),
            - `matplotlib.axes.Axes.contourf()` (bivariate, `fill=True`).

        Returns
        -------
        matplotlib.axes.Axes
            Axes object containing the plot.

        See Also
        --------
        `seaborn.scatterplot <https://seaborn.pydata.org/generated/seaborn.scatterplot.html>`_

        `seaborn.kdeplot <https://seaborn.pydata.org/generated/seaborn.kdeplot.html>`_

        """

        return sspyplot.density(data=self._df, x=x, y=y, incl_scatter=incl_scatter, density_type=density_type,
                                title=title, diagonal_lines=diagonal_lines, xlim=xlim, ylim=ylim,
                                scatter_kws=scatter_kws, incl_outline=incl_outline, figsize=figsize,
                                legend_loc=legend_loc, alpha=alpha, gridsize=gridsize, kernel=kernel, cut=cut,
                                clip=clip, legend=legend, cumulative=cumulative, cbar=cbar, cbar_ax=cbar_ax,
                                cbar_kws=cbar_kws, ax=ax, weights=weights, hue=hue, palette=palette,
                                hue_order=hue_order, hue_norm=hue_norm, multiple=multiple, common_norm=common_norm,
                                common_grid=common_grid, levels=levels, thresh=thresh, bw_method=bw_method,
                                bw_adjust=bw_adjust, log_scale=log_scale, color=color, fill=fill,
                                warn_singular=warn_singular, **kwargs)

    def jointplot(
        self,
        x="ISOPleasant",
        y="ISOEventful",
        incl_scatter=True,
        density_type="full",
        title="Soundscape Joint Plot",
        diagonal_lines=False,
        xlim=(-1, 1),
        ylim=(-1, 1),
        scatter_kws=dict(s=25, linewidth=0),
        incl_outline=False,
        legend_loc="lower left",
        alpha=0.75,
        color=None,
        joint_kws=dict(),
        marginal_kws={"fill": True, "common_norm": False},
        hue=None,
        palette="colorblind",
        hue_order=None,
        hue_norm=None,
        common_norm=False,
        fill=True,
        bw_adjust=None,
        thresh=0.1,
        levels=10,
        legend=False,
        marginal_kind="kde",
    ):  # sourcery skip: default-mutable-arg
        """Create a jointplot with distribution or scatter in the center and distributions on the margins.

        This method works by calling sns.jointplot() and creating a circumplex grid in the joint position, then
        overlaying a density or circumplex_scatter plot. The options for both the joint and marginal plots can be
        passed through the sns.jointplot() separately to customise them separately. The marginal distribution plots
        can be either a density or histogram.

        Parameters
        ----------
        data : pd.DataFrame, np.ndarray, mapping, or sequence
            Input data structure. Either a long-form collection of vectors that can be assigned to named variables or a
            wide-form dataset that will be internally reshaped.
        x : vector or key in `data`, optional
            column name for x variable, by default "ISOPleasant"
        y : vector or key in `data`, optional
            column name for y variable, by default "ISOEventful"
        incl_scatter : bool, optional
            Whether to include a scatter plot of the data, by default True
        density_type : str, optional
            Type of density plot to draw, by default "full"
        diagonal_lines : bool, optional
            whether to include diagonal dimension axis labels in the joint plot, by default False
        palette : str, optional
            [description], by default "colorblind"
        incl_scatter : bool, optional
            plot coordinate scatter underneath density plot, by default False
        fill : bool, optional
            whether to fill the density plot, by default True
        bw_adjust : number, optional
            Factor that multiplicatively scales the value chosen using `bw_method`. Increasing will make the curve smoother.
            by default None
        bw_adjust : [type], optional
            [description], by default default_bw_adjust
        alpha : float, optional
            [description], by default 0.95
        legend : bool, optional
            whether to include the hue labels legend, by default False
        legend_loc : str, optional
            relative location of the legend, by default "lower left"
        marginal_kind : str, optional
            density or histogram plot in the margins, by default "kde"
        hue : vector or key in data, optional
            Grouping variable that will produce points with different colors. Can be either categorical or numeric,
            although color mapping will behave differently in latter case, by default None
        joint_kws : dict, optional
            Arguments to pass to density or scatter joint plot, by default {}
        marginal_kws : dict, optional
            Arguments to pass to marginal distribution plots, by default {"fill": True}
        hue : vector or key in `data`, optional
            Semantic variable that is mapped to determine the color of plot elements.
        palette : string, list, dict, or `matplotlib.colors.Colormap`, optional
            Method for choosing the colors to use when mapping the `hue` semantic. String values are passed to
            `color_palette()`. List or dict values imply categorical mapping, while a colormap object implies numeric
            mapping.
            by default, `"colorblind"`
        hue_order : vector of strings, optional.
            Specify the order of processing and plotting for categorical levels of the `hue` semantic.
        hue_norm : tuple or matplotlib.colors.Normalize, optional
            Either a pair of values that set the normalization range in data units or an object that will map from data
            units into a [0, 1] interval. Usage implies numeric mapping.
        common_norm : bool, optional
            If True, scale each conditional density by the number of observations such that the total area under all
            densities sums to 1. Otherwise, normalize each density independently, by default False.
        fill : bool, optional
            If True, fill in the area under univariate density curves or between bivariate contours. If None, the default
            depends on `multiple`. by default True
        bw_adjust : number, optional
            Factor that multiplicatively scales the value chosen using `bw_method`. Increasing will make the curve smoother.
            See Notes. by default default_bw_adjust (1.2)
        thresh : number in [0, 1], optional
            Lowest iso-proportional level at which to draw a contour line. Ignored when `levels` is a vector. Only relevant
            with bivariate plots. by default 0.1
        levels : int or vector, optional
            Number of contour levels or values to draw contours at. A vector argument must have increasing values in [0, 1].
            Levels correspond to iso-proportionas of the density: e.g. 20% of the probability mass will lie below the
            contour drawn for 0.2. Only relevant with bivariate data.
            by default 10
        legend : bool, optional
            If False, suppress the legend for semantic variables, by default False
        legend_loc : str, optional
            Relative location of the legend, by default "lower left"
        marginal_kind : str, optional
            density or histogram plot in the margins, by default "kde"

        Returns
        -------
        plt.Axes

        See Also
        --------
        `seaborn.kdeplot <https://seaborn.pydata.org/generated/seaborn.kdeplot.html>`_

        `seaborn.jointplot <https://seaborn.pydata.org/generated/seaborn.jointplot.html>`_
        """

        return sspyplot.jointplot(
            data=self._df,
            x=x,
            y=y,
            incl_scatter=incl_scatter,
            density_type=density_type,
            title=title,
            diagonal_lines=diagonal_lines,
            xlim=xlim,
            ylim=ylim,
            scatter_kws=scatter_kws,
            incl_outline=incl_outline,
            legend_loc=legend_loc,
            alpha=alpha,
            color=color,
            joint_kws=joint_kws,
            marginal_kws=marginal_kws,
            hue=hue,
            palette=palette,
            hue_order=hue_order,
            hue_norm=hue_norm,
            common_norm=common_norm,
            fill=fill,
            bw_adjust=bw_adjust,
            thresh=thresh,
            levels=levels,
            legend=legend,
            marginal_kind=marginal_kind,
        )


# %%



# Customized functions specifically for the International Soundscape Database

from datetime import date

# Add soundscapy to the Python path
from pathlib import Path

import pandas as pd
from pandas.api.extensions import register_dataframe_accessor

# Constants and Labels
from soundscapy.utils.surveys import *


@register_dataframe_accessor("isd")
class ISDAccessor:
    """Custom accessor for the International Soundscape Database (ISD) dataset

    Parameters
    ----------
    pandas_obj : pd.DataFrame
        Dataframe containing ISD formatted data

    Attributes
    ----------
    _obj : pd.DataFrame
        Dataframe containing ISD formatted data
    _analysis_data : str
        Date of analysis
    _metadata : dict
        Dictionary of metadata

    Methods
    -------
    validate_dataset
        Validate the dataset to ensure it is in the correct format.
    paq_data_quality
        Return the data quality of the PAQs
    filter_group_ids
        Filter the dataframe by GroupID
    filter_location_ids
        Filter the dataframe by LocationID
    filter_session_ids
        Filter the dataframe by SessionID
    filter_record_ids
        Filter the dataframe by RecordID
    filter_lockdown
        Filter the dataframe by Lockdown
    convert_group_ids_to_index
        Convert the GroupID column to the index
    return_paqs
        Return the PAQs as a dataframe
    soundscapy_describe
        Return a summary of the data
    mean_responses
        Calculate the mean responses for each PAQ
    convert_column_to_index
        Reassign an existing column as the dataframe index.
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self._analysis_data = date.today().isoformat()
        self._metadata = {}

    def validate_dataset(
        self,
        paq_aliases=PAQ_NAMES,
        allow_lockdown=False,
        allow_paq_na=False,
        verbose=1,
        val_range=(1, 5),
    ):
        """Validate the dataset to ensure it is in the correct format.

        Parameters
        ----------
        paq_aliases : list, optional
            List of aliases for the PAQ, by default PAQ_NAMES
        allow_lockdown : bool, optional
            Allow lockdown, by default False
        allow_paq_na : bool, optional
            Allow PAQ to be NA, by default False
        verbose : int, optional
            Print progress, by default 1
        val_range : tuple, optional
            Range of valid values, by default (1, 5)

        Returns
        -------
        pandas.DataFrame
            Validated dataframe
        """

        return validate_dataset(
            self._obj,
            paq_aliases=paq_aliases,
            allow_lockdown=allow_lockdown,
            allow_paq_na=allow_paq_na,
            verbose=verbose,
            val_range=val_range,
        )

    def paq_data_quality(self, verbose=0):
        """Return the data quality of the PAQs

        Parameters
        ----------
        verbose : int, optional
            Print progress, by default 0

        Returns
        -------
        pandas.DataFrame
        """
        return paq_data_quality(self._obj, verbose=verbose)

    def filter_group_ids(self, group_ids):
        """
        Filter the dataframe by GroupID

        Parameters
        ----------
        group_ids : list
            List of GroupIDs to filter by

        Returns
        -------
        pd.DataFrame
            Filtered dataframe
        """
        return self._obj.sspy.filter("GroupID", group_ids)

    def filter_location_ids(self, location_ids):
        """Filter the dataframe by LocationID

        Parameters
        ----------
        location_ids : list
            List of LocationIDs to filter by

        Returns
        -------
        pd.DataFrame
            Filtered dataframe
        """
        return self._obj.sspy.filter("LocationID", location_ids)

    def filter_session_ids(self, session_ids):
        """Filter the dataframe by SessionID

        Parameters
        ----------
        session_ids : list
            List of SessionIDs to filter by

        Returns
        -------
        pd.DataFrame
            Filtered dataframe
        """
        return self._obj.sspy.filter("SessionID", session_ids)

    def filter_record_ids(self, record_ids):
        """Filter the dataframe by RecordID

        Parameters
        ----------
        record_ids : list
            List of RecordIDs to filter by

        Returns
        -------
        pd.DataFrame
            Filtered dataframe
        """
        return self._obj.sspy.filter("RecordID", record_ids)

    def filter_lockdown(self, is_lockdown=False):
        """Filter the dataframe by Lockdown

        Parameters
        ----------
        is_lockdown : bool, optional
            Filter by lockdown, by default False

        Returns
        -------
        pd.DataFrame
            Filtered dataframe
        """
        return (
            self._obj.query("Lockdown == 1")
            if is_lockdown
            else self._obj.query("Lockdown == 0")
        )

    def convert_group_ids_to_index(self, drop=False):
        """Convert the GroupID column to the index

        Parameters
        ----------
        drop : bool, optional
            Drop the GroupID column, by default False

        Returns
        -------
        pd.DataFrame
            Dataframe with GroupID as index
        """
        return convert_column_to_index(self._obj, "Group_ID", drop=drop)

    def return_paqs(self, incl_ids=True, other_cols=None):
        """Return the PAQs as a dataframe

        Parameters
        ----------
        incl_ids : bool, optional
            Include the IDs, by default True
        other_cols : list, optional
            List of other columns to include, by default None

        Returns
        -------
        pd.DataFrame
            Dataframe containing the PAQs
        """
        return return_paqs(self._obj, incl_ids=incl_ids, other_cols=other_cols)

    def location_describe(
        self, location, type="percent", pl_threshold=0, ev_threshold=0
    ):
        """Return a summary of the data

        Parameters
        ----------
        location : str
            Location to describe
        type : str, optional
            Type of summary, by default "percent"
        pl_threshold : int, optional
            PL threshold, by default 0
        ev_threshold : int, optional
            EV threshold, by default 0

        Returns
        -------
        pd.DataFrame
            Summary of the data
        """
        loc_df = self.filter_location_ids(location_ids=[location])
        count = len(loc_df)
        pl_count = len(loc_df[loc_df["ISOPleasant"] > pl_threshold])
        ev_count = len(loc_df[loc_df["ISOEventful"] > ev_threshold])
        vibrant_count = len(
            loc_df.query("ISOPleasant > @pl_threshold & ISOEventful > @ev_threshold")
        )
        chaotic_count = len(
            loc_df.query("ISOPleasant < @pl_threshold & ISOEventful > @ev_threshold")
        )
        mono_count = len(
            loc_df.query("ISOPleasant < @pl_threshold & ISOEventful < @ev_threshold")
        )
        calm_count = len(
            loc_df.query("ISOPleasant > @pl_threshold & ISOEventful < @ev_threshold")
        )

        res = {
            "count": count,
            "ISOPleasant": round(loc_df["ISOPleasant"].mean(), 3),
            "ISOEventful": round(loc_df["ISOEventful"].mean(), 3),
        }
        if type == "percent":
            res["pleasant"] = round(pl_count / count, 3)
            res["eventful"] = round(ev_count / count, 3)
            res["vibrant"] = round(vibrant_count / count, 3)
            res["chaotic"] = round(chaotic_count / count, 3)
            res["monotonous"] = round(mono_count / count, 3)
            res["calm"] = round(calm_count / count, 3)

        elif type == "count":
            res["pleasant"] = pl_count
            res["eventful"] = ev_count
            res["vibrant"] = vibrant_count
            res["chaotic"] = chaotic_count
            res["monotonous"] = mono_count
            res["calm"] = calm_count

        return res

    def soundscapy_describe(self, group_by="LocationID", type="percent"):
        """Return a summary of the data

        Parameters
        ----------
        group_by : str, optional
            Column to group by, by default "LocationID"
        type : str, optional
            Type of summary, by default "percent"

        Returns
        -------
        pd.DataFrame
            Summary of the data
        """
        res = {
            location: self.location_describe(location, type=type)
            for location in self._obj[group_by].unique()
        }

        res = pd.DataFrame.from_dict(res, orient="index")
        return res

    def mean_responses(self, group="LocationID") -> pd.DataFrame:
        """Calculate the mean responses for each PAQ

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe containing ISD formatted data

        Returns
        -------
        pd.Dataframe
            Dataframe containing the mean responses for each PAQ
        """
        return mean_responses(self._obj, group=group)

    def convert_column_to_index(self, col="GroupID", drop=True):
        """Reassign an existing column as the dataframe index.

        Parameters
        ----------
        col : str, optional
            Column to convert to index, by default "GroupID"
        drop : bool, optional
            Drop the column, by default True

        Returns
        -------
        pd.DataFrame
            Dataframe with the column as the index
        """
        return convert_column_to_index(self._obj, col=col, drop=drop)

    def validate_isd(
        self,
        paq_aliases=None,
        allow_lockdown=False,
        allow_paq_na=False,
        verbose=1,
        val_range=(1, 5),
    ):
        """Validate the ISD dataset.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe containing ISD formatted data
        paq_aliases : dict, optional
            Dictionary of PAQ aliases, by default None
        allow_lockdown : bool, optional
            Allow lockdown PAQs?, by default False
        allow_paq_na : bool, optional
            Allow missing PAQs?, by default False
        verbose : int, optional
            Verbosity level, by default 1
        val_range : tuple, optional
            Range of valid values, by default (1,5)

        Returns
        -------
        pd.DataFrame
            Dataframe containing the validation results
        """
        return validate_dataset(
            self._obj,
            paq_aliases=paq_aliases,
            allow_lockdown=allow_lockdown,
            allow_paq_na=allow_paq_na,
            verbose=verbose,
            val_range=val_range,
        )


def load_isd_dataset(version="latest"):
    """Automatically fetch and load the ISD dataset from Zenodo

    Parameters
    ----------
    version : str, optional
        version number of the dataset to fetch, by default "latest"

    Returns
    -------
    pd.Dataframe
        ISD data
    """
    if version.lower() not in ["latest", "v0.2.3", "v0.2.2", "v0.2.1", "v0.2.0"]:
        raise ValueError(f"Version {version} not recognised.")

    version = "v0.2.3" if version == "latest" else version
    if version in ["V0.2.0", "v0.2.0", "V0.2.1", "v0.2.1"]:
        url = "https://zenodo.org/record/5578573/files/SSID%20Lockdown%20Database%20VL0.2.1.xlsx"

    elif version in ["V0.2.2", "v0.2.2"]:
        url = "https://zenodo.org/record/5705908/files/SSID%20Lockdown%20Database%20VL0.2.2.xlsx"

    elif version in ["v0.2.3", "V0.2.3"]:
        url = "https://zenodo.org/record/5914762/files/SSID%20Lockdown%20Database%20VL0.2.2.xlsx"

    return pd.read_excel(url, engine="openpyxl")


# Dealing with Directories!
def collect_all_dirs(
    root_directory: Path,
    location_ids: list,
    param_list: list,
    include_TS: bool = True,
    include_spectrum: bool = True,
    include_WAV: bool = True,
):
    """Iterate throughout the SSID DB file structure to extract TimeSeries, SpectrumData, WAV directory paths.

    Parameters
    ----------
    root_directory : Path
        The city-level SSID directory
    location_ids : list
        A subset of LocationIDs to include in the filepath collection
    param_list : list
        A subset of parameters to include in the filepath collection
    include_TS : bool, optional
        Collect TimeSeries dirs?, by default True
    include_spectrum : bool, optional
        Collect SpectrumData dirs?, by default True
    include_WAV : bool, optional
        Collect WAV dirs?, by default True

    Returns
    -------
    tuple of lists
        A tuple containing the full lists of TimeSeries, SpectrumData, and WAV directories.
        These lists contain WindowsPath objects.

    Examples
    ________
    >>> full_ts_list, full_spectrum_list, full_wav_list = collect_all_dirs(TEST_DIR, ["LocationA", "LocationB"], PARAM_LIST)
    >>> len(full_ts_list)
    33
    >>> len(full_spectrum_list)
    6
    >>> len(full_wav_list)
    3
    """

    # Input tests
    if not isinstance(root_directory, Path):
        raise TypeError(
            "The directory should be provided as a WindowsPath object from pathlib."
        )
    if not root_directory.is_dir():
        raise FileNotFoundError("Path does not exist.")

    # Collect list of session id directories
    session_dirs = [session for session in root_directory.iterdir() if session.is_dir()]
    session_dirs = [session for session in session_dirs if "OFF" in session.name]

    new_session_dirs = [
        [session for location in location_ids if (location in session.name)]
        for session in session_dirs
    ]

    session_dirs = [
        val for sublist in new_session_dirs for val in sublist
    ]  # Remove blank entries

    bin_dirs = []
    for session in session_dirs:
        bin_dir = [
            child
            for child in session.iterdir()
            if child.is_dir() and "BIN" in child.name
        ][0]
        bin_dirs.append(bin_dir)

    full_ts_list = _ts_dirs(bin_dirs, param_list) if include_TS else []
    full_spectrum_list = _spectrum_dirs(bin_dirs) if include_spectrum else []
    full_wav_list = _wav_dirs(bin_dirs) if include_WAV else []
    return full_ts_list, full_spectrum_list, full_wav_list


def _spectrum_dirs(bin_dirs):
    spectrum_dirs = []
    for directory in bin_dirs:
        spectrum_dir = [
            child
            for child in directory.iterdir()
            if child.is_dir() and "SpectrumData" in child.name
        ][0]
        spectrum_dirs.append(spectrum_dir)

    full_spectrum_list = []
    for directory in spectrum_dirs:
        spectrum_dir = [child for child in directory.iterdir() if child.is_dir()]
        full_spectrum_list.append(spectrum_dir)

    full_spectrum_list = [val for sublist in full_spectrum_list for val in sublist]
    return full_spectrum_list


def _ts_dirs(bin_dirs, param_list):
    # Collect Time Series parameter directories
    ts_dirs = []
    for directory in bin_dirs:
        ts_dir = [
            child
            for child in directory.iterdir()
            if child.is_dir() and "TimeSeries" in child.name
        ][0]
        ts_dirs.append(ts_dir)

    param_dirs = []
    for directory in ts_dirs:
        param_dir = [child for child in directory.iterdir() if child.is_dir()]
        param_dirs.append(param_dir)
    param_dirs = [val for sublist in param_dirs for val in sublist]

    full_ts_list = [
        [directory for param in param_list if (param + "_TS" in directory.name)]
        for directory in param_dirs
    ]

    full_ts_list = [val for sublist in full_ts_list for val in sublist]
    return full_ts_list


def _wav_dirs(bin_dirs):
    # Collect Wav directories
    wav_dirs = []
    for directory in bin_dirs:
        wav_dir = [
            child
            for child in directory.iterdir()
            if child.is_dir() and "WAV" in child.name
        ][0]
        wav_dirs.append(wav_dir)

    return wav_dirs



# %%
import warnings
from typing import Union

import maad
import numpy as np
import pandas as pd
from mosqito.sq_metrics import (
    loudness_zwtv,
    roughness_dw,
    sharpness_din_from_loudness,
    sharpness_din_perseg,
    sharpness_din_tv,
)
from scipy import stats

DEFAULT_LABELS = {
    "LZeq": "LZeq",
    "LCeq": "LCeq",
    "LAeq": "LAeq",
    "Leq": "Leq",
    "SEL": "SEL",
    "loudness_zwtv": "N",
    "roughness_dw": "R",
    "sharpness_din_perseg": "S",
    "sharpness_din_from_loudness": "S",
    "sharpness_din_tv": "S",
}


# %%
# Metrics calculations
def _stat_calcs(
    label: str, ts_array: np.ndarray, res: dict, statistics: Union[list, tuple]
):
    """Process a time series array of metrics into a dictionary of statistics.

    Parameters
    ----------
    label : str
        Label to use for the metric in the results dictionary
    ts_array : np.ndarray
        Time series array of the metric. Must be a 1D array.
        Specific time step is not necessary
    res : dict
        results dictionary to add results to
    statistics : Union
        list of statistics to calculate

    Returns
    -------
    dict
        results dictionary with the calculated statistics added
    """

    for stat in statistics:
        if stat in ("avg", "mean"):
            res[f"{label}_{stat}"] = ts_array.mean()
        elif stat == "max":
            res[f"{label}_{stat}"] = ts_array.max()
        elif stat == "min":
            res[f"{label}_{stat}"] = ts_array.min()
        elif stat == "kurt":
            res[f"{label}_{stat}"] = stats.kurtosis(ts_array)
        elif stat == "skew":
            res[f"{label}_{stat}"] = stats.skew(ts_array)
        else:
            res[f"{label}_{stat}"] = np.percentile(ts_array, 100 - stat)
    return res


# %%
def mosqito_metric_1ch(
    s,
    metric: str,
    statistics: Union[tuple, list] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label=None,
    as_df: bool = False,
    return_time_series: bool = False,
    func_args={},
):
    """Calculating a metric and accompanying statistics from Mosqito.

    Parameters
    ----------
    s : Signal or Binaural (single channel)
        Single channel signal to calculate the sound quality indices for
    metric : {"loudness_zwtv", "sharpness_din_from_loudness", "sharpness_din_perseg", "sharpness_din_tv",
    "roughness_zwtv"}
        Metric to calculate
    statistics : tuple or list
        List of level statistics to calculate (e.g. L_5, L_90, etc.),
            by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
    label : str, optional
        Label to use for the metric in the results dictionary, by default None
        If None, will pull from default label for that metric given in DEFAULT_LABELS
    as_df : bool, optional
        Return the results as a dataframe, by default False
    return_time_series : bool, optional
        Return the time series array of the metric, by default False
        Only works for metrics that return a time series array.
        Cannot be returned in a dataframe. Will raise a warning if both `as_df`
        and `return_time_series` are True and will only return the DataFrame with the other stats.
    verbose : bool, optional
        Whether to print status updates, by default False
    **func_args : dict, optional
        Additional keyword arguments to pass to the metric function, by default {}

    Returns
    -------
    dict
        dictionary of the calculated statistics.
        key is metric name + statistic (e.g. LZeq_5, LZeq_90, etc)
        value is the calculated statistic

    Raises
    ------
    ValueError
        Signal must be single channel. Can be a slice of a multichannel signal.
    ValueError
        Metric is not recognized. Must be one of {"loudness_zwtv", "sharpness_din_from_loudness",
        "sharpness_din_perseg", "roughness_dw"}
    Warning

    """
    # Checks and warnings
    if s.channels != 1:
        raise ValueError("Signal must be single channel")
    try:
        label = label or DEFAULT_LABELS[metric]
    except KeyError as e:
        raise ValueError(f"Metric {metric} not recognized.") from e
    if as_df and return_time_series:
        warnings.warn(
            "Cannot return both a dataframe and time series. Returning dataframe only."
        )
        return_time_series = False

    # Start the calc
    res = {}
    if metric == "loudness_zwtv":
        N, N_spec, bark_axis, time_axis = loudness_zwtv(s, s.fs, **func_args)
        res = _stat_calcs(label, N, res, statistics)
        if return_time_series:
            res[f"{label}_ts"] = (time_axis, N)
    elif metric == "roughness_dw":
        R, R_spec, bark_axis, time_axis = roughness_dw(s, s.fs, **func_args)
        res = _stat_calcs(label, R, res, statistics)
        if return_time_series:
            res[f"{label}_ts"] = (time_axis, R)
    elif metric == "sharpness_din_from_loudness":
        # The `sharpness_din_from_loudness` requires the loudness to be calculated first.
        field_type = func_args.get("field_type", "free")
        N, N_spec, bark_axis, time_axis = loudness_zwtv(s, s.fs, field_type=field_type)
        res = _stat_calcs("N", N, res, statistics)
        if return_time_series:
            res["N_ts"] = time_axis, N

        # Calculate the sharpness_din_from_loudness metric
        func_args.pop("field_type", None)
        S = sharpness_din_from_loudness(N, N_spec, **func_args)
        res = _stat_calcs(label, S, res, statistics)
        if return_time_series:
            res[f"{label}_ts"] = (time_axis, S)
    elif metric == "sharpness_din_perseg":
        S, time_axis = sharpness_din_perseg(s, s.fs, **func_args)
        res = _stat_calcs(label, S, res, statistics)
        if return_time_series:
            res[f"{label}_ts"] = (time_axis, S)
    elif metric == "sharpness_din_tv":
        S, time_axis = sharpness_din_tv(s, s.fs, **func_args)
        res = _stat_calcs(label, S, res, statistics)
        if return_time_series:
            res[f"{label}_ts"] = (time_axis, S)
    else:
        raise ValueError(f"Metric {metric} not recognized.")

    # Return the results in the requested format
    if not as_df:
        return res
    try:
        rec = s.recording
        return pd.DataFrame(res, index=[rec])
    except AttributeError:
        return pd.DataFrame(res, index=[0])


# %%
def maad_metric_1ch(
    s, metric: str, as_df: bool = False, verbose: bool = False, func_args={}
):
    """Run a metric from the scikit-maad library (or suite of indices) on a single channel signal.

    Currently only supports running all of the alpha indices at once.

    Parameters
    ----------
    s : Signal or Binaural (single channel)
        Single channel signal to calculate the alpha indices for
    metric : {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}
        Metric to calculate
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    verbose : bool, optional
        Whether to print status updates, by default False
    **func_args : dict, optional
        Additional keyword arguments to pass to the metric function, by default {}

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame
    """
    # Checks and status
    if s.channels != 1:
        raise ValueError("Signal must be single channel")
    if verbose:
        print(f" - Calculating scikit-maad {metric}")
    # Start the calc
    if metric == "all_spectral_alpha_indices":
        Sxx, tn, fn, ext = maad.sound.spectrogram(
            s, s.fs, **func_args
        )  # spectral requires the spectrogram
        res = maad.features.all_spectral_alpha_indices(
            Sxx, tn, fn, extent=ext, **func_args
        )[0]

    elif metric == "all_temporal_alpha_indices":
        res = maad.features.all_temporal_alpha_indices(s, s.fs, **func_args)
    else:
        raise ValueError(f"Metric {metric} not recognized.")
    if not as_df:
        return res.to_dict("records")[0]
    try:
        res["Recording"] = s.recording
        res.set_index(["Recording"], inplace=True)
        return res
    except AttributeError:
        return res


# %%
def pyacoustics_metric_1ch(
    s,
    metric: str,
    statistics: Union[list, tuple] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    as_df: bool = False,
    return_time_series: bool = False,
    verbose: bool = False,
    func_args={},
):
    """Run a metric from the pyacoustics library on a single channel object.

    Parameters
    ----------
    s : Signal or Binaural (single channel slice)
        Single channel signal to calculate the metric for
    metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
        The metric to run
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc),
            by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
    label : str, optional
        Label to use for the metric in the results dictionary, by default None
        If None, will pull from default label for that metric given in DEFAULT_LABELS
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False
        Cannot return time series if as_df is True
    verbose : bool, optional
        Whether to print status updates, by default False
    **func_args : dict, optional
        Additional keyword arguments to pass to the metric function, by default {}

    Returns
    -------
    dict
        dictionary of the calculated statistics.
        key is metric name + statistic (e.g. LZeq_5, LZeq_90, etc)
        value is the calculated statistic

    Raises
    ------
    ValueError
        Metric must be one of {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
    """
    if s.channels != 1:
        raise ValueError("Signal must be single channel")
    try:
        label = label or DEFAULT_LABELS[metric]
    except KeyError as e:
        raise ValueError(f"Metric {metric} not recognized.") from e
    if as_df and return_time_series:
        warnings.warn(
            "Cannot return both a dataframe and time series. Returning dataframe only."
        )

        return_time_series = False
    if verbose:
        print(f" - Calculating Python Acoustics: {metric} {statistics}")
    res = {}
    if metric in {"LZeq", "Leq", "LAeq", "LCeq"}:
        if metric in {"LZeq", "Leq"}:
            weighting = "Z"
        elif metric == "LAeq":
            weighting = "A"
        elif metric == "LCeq":
            weighting = "C"
        if "avg" in statistics or "mean" in statistics:
            stat = "avg" if "avg" in statistics else "mean"
            res[f"{label}"] = s.weigh(weighting).leq()
            statistics = list(statistics)
            statistics.remove(stat)
        if len(statistics) > 0:
            res = _stat_calcs(
                label, s.weigh(weighting).levels(**func_args)[1], res, statistics
            )

        if return_time_series:
            res[f"{label}_ts"] = s.weigh(weighting).levels(**func_args)
    elif metric == "SEL":
        res[f"{label}"] = s.sound_exposure_level()
    else:
        raise ValueError(f"Metric {metric} not recognized.")
    if not as_df:
        return res
    try:
        rec = s.recording
        return pd.DataFrame(res, index=[rec])
    except AttributeError:
        return pd.DataFrame(res, index=[0])



"""
Soundscapy is a Python library for soundscape analysis and visualisation.
"""

from soundscapy.analysis._AnalysisSettings import AnalysisSettings, get_default_yaml
from soundscapy.analysis._Binaural import Binaural
from soundscapy.utils import _sspy_accessor, surveys
from soundscapy.databases import isd, araus



"""
The module containing functions for dealing with soundscape survey data.

Notes
-----
The functions in this module are designed to be fairly general and can be used with any dataset in a similar format to
the ISD. The key to this is using a simple dataframe/sheet with the following columns:
    Index columns: e.g. LocationID, RecordID, GroupID, SessionID
    Perceptual attributes: PAQ1, PAQ2, ..., PAQ8
    Independent variables: e.g. Laeq, N5, Sharpness, etc.

The key functions of this module are designed to clean/validate datasets, calculate ISO coordinate values or SSM metrics,
filter on index columns. Functions and operations which are specific to a particular dataset are located in their own
modules under `soundscape.databases`.
"""

from pathlib import Path
from typing import Union, Tuple, List, Dict

import janitor
import numpy as np
import pandas as pd
from scipy import optimize

# Constants and Labels
from soundscapy.utils.parameters import PAQ_IDS, PAQ_NAMES

DEFAULT_CATS = [
    "indexing",
    "sound_source_dominance",
    "raw_PAQs",
    "overall_soundscape",
]

# General helper functions
_flatten = lambda t: [item for sublist in t for item in sublist]


###########
def return_paqs(df, incl_ids=True, other_cols=None):
    """Return only the PAQ columns

    Parameters
    ----------
    incl_ids : bool, optional
        whether to include ID cols too (i.e. RecordID, GroupID, etc), by default True
    other_cols : list, optional
        other columns to also include, by default None

    Returns
    -------
    pd.DataFrame
        dataframe containing only the PAQ columns
    """
    cols = PAQ_IDS
    if incl_ids:
        id_cols = [
            name
            for name in ["RecordID", "GroupID", "SessionID", "LocationID"]
            if name in df.columns
        ]

        cols = id_cols + cols
    if other_cols:
        cols = cols + other_cols
    return df[cols]


def mean_responses(df: pd.DataFrame, group: str) -> pd.DataFrame:
    """Calculate the mean responses for each PAQ

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing ISD formatted data

    Returns
    -------
    pd.Dataframe
        Dataframe containing the mean responses for each PAQ
    """
    df = return_paqs(df, incl_ids=False, other_cols=[group])
    return df.groupby(group).mean()


def _circ_scale(range):
    diff = max(range) - min(range)
    return diff + diff * np.sqrt(2)


def convert_column_to_index(df, col: str, drop=False):
    """Reassign an existing column as the dataframe index"""
    assert col in df.columns, f"col: {col} not found in dataframe"
    df.index = df[col]
    if drop:
        df = df.drop(col, axis=1)
    return df


def validate_dataset(
    df: pd.DataFrame,
    paq_aliases: Union[List, Dict] = None,
    allow_lockdown: bool = False,
    allow_paq_na: bool = False,
    verbose: int = 1,
    val_range: Tuple = (1, 5),
):
    """Performs data quality checks and validates that the dataset fits the expected format

    Parameters
    ----------
    df : pd.DataFrame
        ISD style dataframe, incl PAQ data
    paq_aliases : list or dict, optional
        list of PAQ names (in order)
        or dict of PAQ names with new names as values, by default None
    allow_lockdown : bool, optional
        if True will keep Lockdown data in the df, by default True
    allow_paq_na : bool, optional
        remove rows which have any missing PAQ values
        otherwise will remove those with 50% missing, by default False    verbose : int, optional
        how much info to print while running, by default 1
    val_range : tuple, optional
        min and max range of the PAQ response values, by default (5, 1)

    Returns
    -------
    tuple
        cleaned dataframe, dataframe of excluded samples
    """
    if verbose > 0:
        print("Renaming PAQ columns.")
    df = rename_paqs(df, paq_aliases)

    if verbose > 0:
        print("Checking PAQ data quality.")
    l = paq_data_quality(df, verbose, allow_lockdown, allow_paq_na, val_range)
    if l is None:
        excl_df = None
    else:
        excl_df = df.iloc[l, :]
        df = df.drop(df.index[l])
    return df, excl_df


def rename_paqs(
    df: pd.DataFrame, paq_aliases: Union[Tuple, Dict] = None, verbose: int = 0
) -> pd.DataFrame:
    """
    The rename_paqs function renames the PAQ columns in a dataframe.

    Soundscapy works with PAQ IDs (PAQ1, PAQ2, etc), so if you use labels such as pleasant, vibrant, etc. these will
    need to be renamed.

    It takes as input a pandas DataFrame and returns the same DataFrame with renamed columns.
    If no arguments are passed, it will attempt to rename all of the PAQs based on their column names.

    Parameters
    ----------
        df: pd.DataFrame
            Specify the dataframe to be renamed
        paq_aliases: tuple or dict, optional
            Specify which paqs are to be renamed, by default None.

            If None, will check if the column names are in our pre-defined options (i.e. pleasant, vibrant, etc).

            If a tuple is passed, the order of the tuple must match the order of the PAQs in the dataframe.

            Allow the function to be called with a dictionary of aliases if desired
        verbose: int, optional
            Print out a message if the paqs are already correctly named, by default 0

    Returns
    -------

        A pandas dataframe with the paq_ids column names

    """
    if paq_aliases is None:
        if any(i in b for i in PAQ_IDS for b in df.columns):
            if verbose > 0:
                print("PAQs already correctly named.")
            return df
        if any(i in b for i in PAQ_NAMES for b in df.columns):
            paq_aliases = PAQ_NAMES

    if type(paq_aliases) == list:
        return df.rename(
            columns={
                paq_aliases[0]: PAQ_IDS[0],
                paq_aliases[1]: PAQ_IDS[1],
                paq_aliases[2]: PAQ_IDS[2],
                paq_aliases[3]: PAQ_IDS[3],
                paq_aliases[4]: PAQ_IDS[4],
                paq_aliases[5]: PAQ_IDS[5],
                paq_aliases[6]: PAQ_IDS[6],
                paq_aliases[7]: PAQ_IDS[7],
            }
        )
    elif type(paq_aliases) == dict:
        return df.rename(columns=paq_aliases)


def paq_data_quality(
    df: pd.DataFrame,
    verbose: int = 0,
    allow_lockdown: bool = True,
    allow_na: bool = False,
    val_range: tuple = (1, 5),
) -> Union[List, None]:
    """Basic check of PAQ data quality

    The paq_data_quality function takes a DataFrame and returns a list of indices that
    should be dropped from the DataFrame. The function checks for:

    - Rows with all values equal to 1 (indicating no PAQ data)

    - Rows with more than 4 NaN values (indicating missing PAQ data)

    - Rows where any value is greater than 5 or less than 1 (indicating invalid PAQ data)

    Parameters
    ----------
        df: pd.DataFrame
            Specify the dataframe to be evaluated
        verbose: int, optional
            Determine whether or not the function should print out information about the data quality check, by default 0
        allow_lockdown: bool, optional
            Allow the user to decide whether they want to remove samples that have a lockdown value of 1, by default True
        allow_na: bool
            Ensure that rows with any missing values are dropped, by default False
        val_range: tuple, optional
            Set the range of values that are considered to be valid, by default (1, 5).

    Returns
    -------

        A list of indices that need to be removed from the dataframe

    """

    paqs = return_paqs(df, incl_ids=False)
    l = []
    for i in range(len(paqs)):
        row = paqs.iloc[i]
        if "Lockdown" in df.columns:
            if allow_lockdown and df.iloc[i]["Lockdown"] == 1:
                continue
        if allow_na is False and row.isna().sum() > 0:
            l.append(i)
            continue
        if row["PAQ1"] == row["PAQ2"] == row["PAQ3"] == row["PAQ4"] == row[
            "PAQ5"
        ] == row["PAQ6"] == row["PAQ7"] == row["PAQ8"] and row.sum() != np.mean(
            val_range
        ):
            l.append(i)
        elif row.isna().sum() > 4:
            l.append(i)
        elif row.max() > max(val_range) or row.min() < min(val_range):
            l.append(i)
    if l:
        if verbose > 0:
            print(f"Identified {len(l)} samples to remove.\n{l}")
        return l
    if verbose > 0:
        print("PAQ quality confirmed. No rows dropped.")
    return None


def simulation(n=3000, val_range=(1, 5), add_paq_coords=False, **coord_kwargs):
    """Generate random PAQ responses

    The PAQ responses will follow a uniform random distribution
    for each PAQ, meaning e.g. for calm either 1, 2, 3, 4, or 5
    is equally likely.
    Parameters
    ----------
    n : int, optional
        number of samples to simulate, by default 3000
    add_paq_coords : bool, optional
        should we also calculate the ISO coordinates, by default False

    Returns
    -------
    pd.Dataframe
        dataframe of randomly generated PAQ response
    """
    np.random.seed(42)
    df = pd.DataFrame(
        np.random.randint(min(val_range), max(val_range) + 1, size=(n, 8)),
        columns=PAQ_IDS,
    )
    if add_paq_coords:
        ISOPl, ISOEv = calculate_paq_coords(df, **coord_kwargs)
        df = janitor.add_columns(df, ISOPleasant=ISOPl, ISOEventful=ISOEv)
    return df


def calculate_paq_coords(
    results_df: pd.DataFrame,
    scale_to_one: bool = True,
    val_range: tuple = (5, 1),
    projection: bool = True,
):
    """Calculates the projected ISOPleasant and ISOEventful coordinates

    If a value is missing, by default it is replaced with neutral (3).
    The raw PAQ values should be Likert data from 1 to 5 and the column
    names should match the PAQ_cols given above.

    Parameters
    ----------
    results_df : pd.DataFrame
        Dataframe containing ISD formatted data
    scale_to_one : bool, optional
        Should the coordinates be scaled to (-1, +1), by default True
    projection : bool, optional
        Use the trigonometric projection (cos(45)) term for diagonal PAQs, by default True

    Returns
    -------
    tuple
        ISOPleasant and ISOEventful coordinate values
    """

    proj = np.cos(np.deg2rad(45)) if projection else 1
    scale = _circ_scale(val_range) if scale_to_one else 1

    # TODO: Add if statements for too much missing data
    # P =(p−a)+cos45°(ca−ch)+cos45°(v−m)
    complex_pleasant = (
        (results_df.PAQ1.fillna(3) - results_df.PAQ5.fillna(3))
        + proj * (results_df.PAQ8.fillna(3) - results_df.PAQ4.fillna(3))
        + proj * (results_df.PAQ2.fillna(3) - results_df.PAQ6.fillna(3))
    )
    ISOPleasant = complex_pleasant / scale

    # E =(e−u)+cos45°(ch−ca)+cos45°(v−m)
    complex_eventful = (
        (results_df.PAQ3.fillna(3) - results_df.PAQ7.fillna(3))
        + proj * (results_df.PAQ4.fillna(3) - results_df.PAQ8.fillna(3))
        + proj * (results_df.PAQ2.fillna(3) - results_df.PAQ6.fillna(3))
    )
    ISOEventful = complex_eventful / scale

    return ISOPleasant, ISOEventful


def calculate_polar_coords(results_df: pd.DataFrame, scaling: str = "iso"):
    """Calculates the polar coordinates

    Based on the calculation given in Gurtman and Pincus (2003), pg. 416.

    The raw PAQ values should be Likert data from 1 to 5 and the column
    names should match the PAQ_cols given above.

    Parameters
    ----------
    results_df : pd.DataFrame
        Dataframe containing ISD formatted data
    scaling : str, optional
        The scaling to use for the polar coordinates, by default 'iso'

        Options are 'iso', 'gurtman', and 'none'

        For 'iso', the cartesian coordinates are scaled to (-1, +1) according to the basic
        method given in ISO12913.

        For 'gurtman', the polar coordinates are scaled according to the method given in
        Gurtman and Pincus (2003), pg. 416.

        For 'none', no scaling is applied.

    Returns
    -------
    tuple
        Polar coordinates
    """
    # raise error if scaling is not one of the options
    if scaling not in ["iso", "gurtman", "none"]:
        raise ValueError(
            f"Scaling must be one of 'iso', 'gurtman', or 'none', not {scaling}"
        )

    scale_to_one = True if scaling == "iso" else False
    isopl, isoev = calculate_paq_coords(results_df, scale_to_one=scale_to_one)

    if scaling == "gurtman":
        isopl = isopl * 0.25
        isoev = isoev * 0.25

    r, theta = _convert_to_polar_coords(isopl, isoev)
    return r, theta


def _convert_to_polar_coords(x, y):
    """Convert cartesian coordinates to polar coordinates

    Parameters
    ----------
    x : float
        x coordinate
    y : float
        y coordinate

    Returns
    -------
    tuple
        (r, theta) polar coordinates
    """
    r = np.sqrt(x**2 + y**2)
    theta = np.rad2deg(np.arctan2(y, x))
    return r, theta


def ssm_metrics(
    df: pd.DataFrame,
    paq_cols: list = PAQ_IDS,
    method: str = "cosine",
    val_range: tuple = (5, 1),
    scale_to_one: bool = True,
    angles: Tuple = (0, 45, 90, 135, 180, 225, 270, 315),
):
    """Calculate the SSM metrics for each response

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe containing ISD formatted data
    paq_cols : list, optional
        List of PAQ columns, by default PAQ_IDS
    method : str, optional
        Method by which to calculate the SSM, by default 'cosine'
        'cosine' fits a cosine model to the data, using the Structural Summary Method developed
        by Gurtman (1994; Gurtman & Balakrishnan, 1998).
        'polar_conversion' directly converts the ISO coordinates to polar coordinates.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the SSM metrics
    """
    # Check that the PAQ columns are present
    if not set(paq_cols).issubset(df.columns):
        raise ValueError("PAQ columns are not present in the dataframe.")

    # # Check that the PAQ values are within the range
    # if not _check_paq_range(df, paq_cols, val_range, verbose):
    #     raise ValueError("PAQ values are not within the specified range.")

    if method == "polar":
        # Calculate the coordinates
        vl, theta = calculate_polar_coords(df)

        mean = np.mean(df[paq_cols], axis=1)
        mean = mean / abs(max(val_range) - min(val_range)) if scale_to_one else mean

        # Calculate the SSM metrics
        df = janitor.add_columns(
            df,
            vl=vl,
            theta=theta,
            mean_level=mean,
        )
        return df

    elif method == "cosine":

        ssm_df = df[paq_cols].apply(
            lambda y: ssm_cosine_fit(y, angles=angles), axis=1, result_type="expand"
        )

        df = janitor.add_columns(
            df,
            amp=ssm_df.iloc[:, 0],
            delta=ssm_df.iloc[:, 1],
            elev=ssm_df.iloc[:, 2],
            dev=ssm_df.iloc[:, 3],
            r2=ssm_df.iloc[:, 4],
        )
        return df

    else:
        raise ValueError("Method must be either 'polar' or 'cosine'.")


def ssm_cosine_fit(
    y,
    angles=(0, 45, 90, 135, 180, 225, 270, 315),
    bounds=([0, 0, 0, -np.inf], [np.inf, 360, np.inf, np.inf]),
):
    """Fit a cosine model to the data

    Parameters
    ----------
    angles : list
        List of angles
    y : list
        List of y values
    bounds : tuple
        Bounds for the parameters

    Returns
    -------
    tuple
        (amp, delta, elev, dev)
    """

    def form(theta, amp, delta, elev, dev):
        return elev + amp * np.cos(np.radians(theta - delta)) + dev

    param, covariance = optimize.curve_fit(
        form,
        xdata=angles,
        ydata=y,
        bounds=bounds,
    )
    r2 = _r2_score(y, form(angles, *param))
    amp, delta, elev, dev = param
    return amp, delta, elev, dev, r2


def _r2_score(y, y_hat):
    """Calculates the R2 score

    Parameters
    ----------
    y : np.ndarray
        Actual values
    y_hat : np.ndarray
        Predicted values

    Returns
    -------
    float
        R2 score
    """
    y_bar = np.mean(y)
    ss_tot = np.sum((y - y_bar) ** 2)
    ss_res = np.sum((y - y_hat) ** 2)
    return 1 - (ss_res / ss_tot)


# %%
if __name__ == "__main__":
    import doctest

    TEST_DIR = Path("../../soundscapy/test/test_DB")
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



from pathlib import Path
from time import localtime, strftime
from typing import Union
import yaml
import urllib.request


def get_default_yaml(save_as="default_settings.yaml"):
    """
    Retrieves the default settings for analysis from the GitHub repository
    and saves them to a file.

    Parameters
    ----------
    save_as : str, optional
        The name of the file to save the default settings to. Defaults to
        "default_settings.yaml".
    """
    print("Downloading default settings from GitHub...")
    urllib.request.urlretrieve(
        "https://raw.githubusercontent.com/MitchellAcoustics/Soundscapy/main/soundscapy/analysis/default_settings.yaml",
        save_as,
    )


class AnalysisSettings(dict):
    """Dict of settings for analysis methods. Each library has a dict of metrics,
    each of which has a dict of settings.
    """

    def __init__(
        self,
        data,
        run_stats=True,
        force_run_all=False,
        filepath: Union[str, Path] = None,
    ):
        super().__init__(data)
        self.run_stats = run_stats
        self.force_run_all = force_run_all
        self.filepath = filepath
        runtime = strftime("%Y-%m-%d %H:%M:%S", localtime())
        super().__setitem__("runtime", runtime)

    @classmethod
    def from_yaml(cls, filename: Union[Path, str], run_stats=True, force_run_all=False):
        """Generate a settings object from a yaml file.

        Parameters
        ----------
        filename : Path object or str
            filename of the yaml file
        run_stats : bool, optional
            whether to include all stats listed or just return the main metric, by default True

            This can simplify the results dataframe if you only want the main metric.
            For example, rather than including L_5, L_50, etc. will only include LEq
        force_run_all : bool, optional
            whether to force all metrics to run regardless of what is set in their options, by default False

            Use Cautiously. This can be useful if you want to run all metrics, but don't want to change the yaml file.
            Warning: If both mosqito:loudness_zwtv and mosqitsharpness_din_from_loudness are present in the settings file, this will result in the loudness calc being run twice.
        Returns
        -------
        AnalysisSettings
            AnalysisSettings object
        """
        with open(filename, "r") as f:
            return cls(yaml.load(f, Loader=yaml.Loader), run_stats, force_run_all, filename)

    @classmethod
    def default(cls, run_stats=True, force_run_all=False):
        """Generate a default settings object.

        Parameters
        ----------
        run_stats : bool, optional
            whether to include all stats listed or just return the main metric, by default True

            This can simplify the results dataframe if you only want the main metric.
            For example, rather than including L_5, L_50, etc. will only include LEq
        force_run_all : bool, optional
            whether to force all metrics to run regardless of what is set in their options, by default False

            Use Cautiously. This can be useful if you want to run all metrics, but don't want to change the yaml file.
            Warning: If both mosqito:loudness_zwtv and mosqito:sharpness_din_from_loudness are present in the settings
            file, this will result in the loudness calc being run twice.

        Returns
        -------
        AnalysisSettings
            AnalysisSettings object
        """
        import soundscapy

        root = Path(soundscapy.__path__[0])
        return cls(
            AnalysisSettings.from_yaml(
                Path(root, "analysis", "default_settings.yaml"),
                run_stats,
                force_run_all,
            )
        )

    def reload(self):
        """Reload the settings from the yaml file."""
        return self.from_yaml(self.filepath, self.run_stats, self.force_run_all)

    def to_yaml(self, filename: Union[Path, str]):
        """Save settings to a yaml file.

        Parameters
        ----------
        filename : Path object or str
            filename of the yaml file
        """
        with open(filename, "w") as f:
            yaml.dump(self, f)

    def parse_maad_all_alpha_indices(self, metric: str):
        """Generate relevant settings for the maad all_alpha_indices methods.

        Parameters
        ----------
        metric : str
            metric to prepare for

        Returns
        -------
        run: bool
            Whether to run the metric
        channel: tuple or list of str, or str
            channel(s) to run the metric on
        """
        lib_settings = self["scikit-maad"].copy()
        run = lib_settings[metric]["run"] or self.force_run_all
        channel = lib_settings[metric]["channel"].copy()
        return run, channel

    def parse_pyacoustics(self, metric: str):
        """Generate relevant settings for a pyacoustics metric.

        Parameters
        ----------
        metric : str
            metric to prepare for

        Returns
        -------
        run: bool
            Whether to run the metric
        channel: tuple or list of str, or str
            channel(s) to run the metric on
        statistics: tuple or list of str, or str
            statistics to run the metric on.
            If run_stats is False, will only return the main statistic
        label: str
            label to use for the metric
        func_args: dict
            arguments to pass to the underlying metric function from python acoustics
        """
        return self._parse_method("PythonAcoustics", metric)

    def parse_mosqito(self, metric: str):
        """Generate relevant settings for a mosqito metric.

        Parameters
        ----------
        metric : str
            metric to prepare for

        Returns
        -------
        run: bool
            Whether to run the metric
        channel: tuple or list of str, or str
            channel(s) to run the metric on
        statistics: tuple or list of str, or str
            statistics to run the metric on.
            If run_stats is False, will only return the main statistic
        label: str
            label to use for the metric
        func_args: dict
            arguments to pass to the underlying metric function from MoSQITo
        """
        run, channel, statistics, label, func_args = self._parse_method(
            "MoSQITo", metric
        )
        try:
            parallel = self["MoSQITo"][metric]["parallel"]
        except KeyError:
            parallel = False
        # Check for sub metric
        # if sub metric is present, don't run this metric
        if (
            metric == "loudness_zwtv"
            and "sharpness_din_from_loudness" in self["MoSQITo"].keys()
            and self["MoSQITo"]["sharpness_din_from_loudness"]["run"]
            and self.force_run_all is False
        ):
            run = False
        return run, channel, statistics, label, parallel, func_args

    def _parse_method(self, library: str, metric: str):
        """Helper function to return relevant settings for a library

        Parameters
        ----------
        library : str
            Library containing the metric
        metric : str
            metric to prepare for

        Returns
        -------
        run: bool
            Whether to run the metric
        channel: tuple or list of str, or str
            channel(s) to run the metric on
        statistics: tuple or list of str, or str
            statistics to run the metric on.
            If run_stats is False, will only return the main statistic
        label: str
            label to use for the metric
        func_args: dict
            arguments to pass to the underlying metric function from the library
        """
        lib_settings = self[library].copy()
        channel = lib_settings[metric]["channel"].copy()
        statistics = lib_settings[metric]["statistics"].copy() if self.run_stats else []
        label = lib_settings[metric]["label"]
        func_args = (
            lib_settings[metric]["func_args"]
            if "func_args" in lib_settings[metric].keys()
            else {}
        )

        main_stat = lib_settings[metric]["main"]
        statistics = (
            (list(statistics) + [main_stat])
            if main_stat not in statistics
            else tuple(statistics)
        )

        # Override metric run settings if force_run_all is True
        run = lib_settings[metric]["run"] or self.force_run_all
        return run, channel, statistics, label, func_args


# %%



from pathlib import Path

from acoustics import Signal

from soundscapy import AnalysisSettings
from soundscapy.analysis.binaural import *
from soundscapy.analysis.metrics import *


class Binaural(Signal):
    """Binaural signal class for analysis of binaural signals.
    A signal consisting of 2D samples (array) and a sampling frequency (fs).

    Subclasses the Signal class from python acoustics.
    Also adds attributes for the recording name.
    Adds the ability to do binaural analysis using the acoustics, scikit-maad and mosqito libraries.
    Optimised for batch processing with analysis settings predefined in a yaml file and passed to the class
    via the AnalysisSettings class.

    See Also
    --------
    acoustics.Signal : Base class for binaural signal
    """

    def __new__(cls, data, fs, recording="Rec"):
        obj = super().__new__(cls, data, fs).view(cls)
        obj.recording = recording
        if obj.channels != 2:
            raise ValueError("Binaural class only supports 2 channels.")
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.fs = getattr(obj, "fs", None)
        self.recording = getattr(obj, "recording", None)

    def calibrate_to(self, decibel: Union[float, list, tuple], inplace: bool = False):
        """Calibrate two channel signal to predefined Leq/dB levels.

        Parameters
        ----------
        decibel : float, list or tuple of float
            Value(s) to calibrate to in dB (Leq)
            Can also handle np.ndarray and pd.Series of length 2.
            If only one value is passed, will calibrate both channels to the same value.
        inplace : bool, optional
            Whether to perform inplace or not, by default False

        Returns
        -------
        Binaural
            Calibrated Binaural signal

        Raises
        ------
        ValueError
            If decibel is not a (float, int) or a list or tuple of length 2.

        See Also
        --------
        acoustics.Signal.calibrate_to : Base method for calibration. Cannot handle 2ch calibration
        """
        if isinstance(decibel, (np.ndarray, pd.Series)):  # Force into tuple
            decibel = tuple(decibel)
        if isinstance(decibel, (list, tuple)):
            if len(decibel) == 2:  # Per-channel calibration (recommended)
                decibel = np.array(decibel)
                decibel = decibel[..., None]
                return super().calibrate_to(decibel, inplace)
            elif (
                len(decibel) == 1
            ):  # if one value given in tuple, assume same for both channels
                decibel = decibel[0]
            else:
                raise ValueError(
                    "decibel must either be a single value or a 2 value tuple"
                )
        if isinstance(decibel, (int, float)):  # Calibrate both channels to same value
            return super().calibrate_to(decibel, inplace)
        else:
            raise ValueError("decibel must be a single value or a 2 value tuple")

    @classmethod
    def from_wav(
        cls,
        filename: Union[Path, str],
        calibrate_to: Union[float, list, tuple] = None,
        normalize: bool = False,
    ):
        """Load a wav file and return a Binaural object

        Overrides the Signal.from_wav method to return a
        Binaural object instead of a Signal object.

        Parameters
        ----------
        filename : Path, str
            Filename of wav file to load
        calibrate_to : float, list or tuple of float, optional
            Value(s) to calibrate to in dB (Leq)
            Can also handle np.ndarray and pd.Series of length 2.
            If only one value is passed, will calibrate both channels to the same value.
        normalize : bool, optional
            Whether to normalize the signal, by default False

        Returns
        -------
        Binaural
            Binaural signal object of wav recording

        See Also
        --------
        acoustics.Signal.from_wav : Base method for loading wav files
        """
        s = super().from_wav(filename, normalize)
        if calibrate_to is not None:
            s.calibrate_to(calibrate_to, inplace=True)
        return cls(s, s.fs, recording=filename.stem)

    def _get_channel(self, channel):
        """Get a single channel from the signal

        Parameters
        ----------
        channel : int
            Channel to get (0 or 1)

        Returns
        -------
        Signal
            Single channel signal
        """
        if self.channels == 1:
            return self
        elif (
            channel is None
            or channel == "both"
            or channel == ("Left", "Right")
            or channel == ["Left", "Right"]
        ):
            return self
        elif channel in ["Left", 0, "L"]:
            return self[0]
        elif channel in ["Right", 1, "R"]:
            return self[1]
        else:
            warnings.warn("Channel not recognised. Returning Binaural object as is.")
            return self

    # Python Acoustics metrics
    def pyacoustics_metric(
        self,
        metric: str,
        statistics: Union[tuple, list] = (
            5,
            10,
            50,
            90,
            95,
            "avg",
            "max",
            "min",
            "kurt",
            "skew",
        ),
        label: str = None,
        channel: Union[str, int, list, tuple] = ("Left", "Right"),
        as_df: bool = True,
        return_time_series: bool = False,
        verbose: bool = False,
        analysis_settings: AnalysisSettings = None,
        func_args={},
    ):
        """Run a metric from the python acoustics library

        Parameters
        ----------
        metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
            The metric to run.
        statistics : tuple or list, optional
            List of level statistics to calulate (e.g. L_5, L_90, etc.),
                by default ( 5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew", )
        label : str, optional
            Label to use for the metric, by default None
            If None, will pull from default label for that metric given in sq_metrics.DEFAULT_LABELS
        channel : tuple, list, or str, optional
            Which channels to process, by default None
            If None, will process both channels
        as_df: bool, optional
            Whether to return a dataframe or not, by default True
            If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
        return_time_series: bool, optional
            Whether to return the time series of the metric, by default False
            Cannot return time series if as_df is True
        verbose : bool, optional
            Whether to print status updates, by default False
        analysis_settings : AnalysisSettings, optional
            Settings for analysis, by default None

            Any settings given here will override those in the other options.
            Can pass any *args or **kwargs to the underlying python acoustics method.
        Returns
        -------
        dict or pd.DataFrame
            Dictionary of results if as_df is False, otherwise a pandas DataFrame

        See Also
        --------
        metrics.pyacoustics_metric
        acoustics.standards_iso_tr_25417_2007.equivalent_sound_pressure_level : Base method for Leq calculation
        acoustics.standards.iec_61672_1_2013.sound_exposure_level : Base method for SEL calculation
        acoustics.standards.iec_61672_1_2013.time_weighted_sound_level : Base method for Leq level time series calculation
        """
        if analysis_settings:
            (
                run,
                channel,
                statistics,
                label,
                func_args,
            ) = analysis_settings.parse_pyacoustics(metric)
            if run is False:
                return None

        channel = ("Left", "Right") if channel is None else channel
        s = self._get_channel(channel)

        if s.channels == 1:
            return pyacoustics_metric_1ch(
                s,
                metric,
                statistics,
                label,
                as_df,
                return_time_series,
                verbose,
                func_args,
            )

        else:
            return pyacoustics_metric_2ch(
                s,
                metric,
                statistics,
                label,
                channel,
                as_df,
                return_time_series,
                verbose,
                func_args,
            )

    # # Mosqito Metrics
    def mosqito_metric(
        self,
        metric: str,
        statistics: Union[tuple, list] = (
            5,
            10,
            50,
            90,
            95,
            "avg",
            "max",
            "min",
            "kurt",
            "skew",
        ),
        label: str = None,
        channel: Union[int, tuple, list, str] = ("Left", "Right"),
        as_df: bool = True,
        return_time_series: bool = False,
        parallel: bool = True,
        verbose: bool = False,
        analysis_settings: AnalysisSettings = None,
        func_args={},
    ):
        """Run a metric from the mosqito library

        Parameters
        ----------
        metric : {"loudness_zwtv", "sharpness_din_from_loudness", "sharpness_din_perseg",
        "sharpness_tv", "roughness_dw"}
            Metric to run from mosqito library.

            In the case of "sharpness_din_from_loudness", the "loudness_zwtv" metric
            will be calculated first and then the sharpness will be calculated from that.
            This is because the sharpness_from loudness metric requires the loudness metric to be
            calculated. Loudness will be returned as well
        statistics : tuple or list, optional
            List of level statistics to calculate (e.g. L_5, L_90, etc.),
                by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
        label : str, optional
            Label to use for the metric, by default None
            If None, will pull from default label for that metric given in sq_metrics.DEFAULT_LABELS
        channel : tuple or list of str or str, optional
            Which channels to process, by default ("Left", "Right")
        as_df: bool, optional
            Whether to return a dataframe or not, by default True
            If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
        return_time_series: bool, optional
            Whether to return the time series of the metric, by default False
            Cannot return time series if as_df is True
        parallel : bool, optional
            Whether to run the channels in parallel, by default True
            If False, will run each channel sequentially.
            If being run as part of a larger parallel analysis (e.g. processing many recordings at once), this will
            automatically be set to False.
        verbose : bool, optional
            Whether to print status updates, by default False
        analysis_settings : AnalysisSettings, optional
            Settings for analysis, by default None

            Any settings given here will override those in the other options.
            Can pass any *args or **kwargs to the underlying python acoustics method.
        Returns
        -------
        dict or pd.DataFrame
            Dictionary of results if as_df is False, otherwise a pandas DataFrame

        See Also
        --------
        binaural.mosqito_metric_2ch : Method for running metrics on 2 channels
        binaural.mosqito_metric_1ch : Method for running metrics on 1 channel
        """
        if analysis_settings:
            (
                run,
                channel,
                statistics,
                label,
                parallel,
                func_args,
            ) = analysis_settings.parse_mosqito(metric)
            if run is False:
                return None

        channel = ("Left", "Right") if channel is None else channel
        s = self._get_channel(channel)

        if s.channels == 1:
            return mosqito_metric_1ch(
                s, metric, statistics, label, as_df, return_time_series, func_args
            )
        else:
            return mosqito_metric_2ch(
                s,
                metric,
                statistics,
                label,
                channel,
                as_df,
                return_time_series,
                parallel,
                verbose,
                func_args,
            )

    # scikit-maad metrics
    def maad_metric(
        self,
        metric: str,
        channel: Union[int, tuple, list, str] = ("Left", "Right"),
        as_df: bool = True,
        verbose: bool = False,
        analysis_settings: AnalysisSettings = None,
        func_args={},
    ):
        """Run a metric from the scikit-maad library

        Currently only supports running all of the alpha indices at once.

        Parameters
        ----------
        metric : {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}
            The metric to run
        channel : tuple, list or str, optional
            Which channels to process, by default None
        as_df: bool, optional
            Whether to return a dataframe or not, by default True
            If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
        verbose : bool, optional
            Whether to print status updates, by default False
        analysis_settings : AnalysisSettings, optional
            Settings for analysis, by default None

            Any settings given here will override those in the other options.
            Can pass any *args or **kwargs to the underlying python acoustics method.
        Returns
        -------
        dict or pd.DataFrame
            Dictionary of results if as_df is False, otherwise a pandas DataFrame


        Raises
        ------
        ValueError
            If metric name is not recognised.
        """
        if analysis_settings:
            if metric in {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}:
                run, channel = analysis_settings.parse_maad_all_alpha_indices(metric)
            else:
                raise ValueError(f"Metric {metric} not recognised")
            if run is False:
                return None
        channel = ("Left", "Right") if channel is None else channel
        s = self._get_channel(channel)
        if s.channels == 1:
            return maad_metric_1ch(s, metric, as_df, verbose, func_args)
        else:
            return maad_metric_2ch(s, metric, channel, as_df, verbose, func_args)

    def process_all_metrics(
        self,
        analysis_settings: AnalysisSettings,
        parallel: bool = True,
        verbose: bool = False,
    ):
        """Run all metrics specified in the AnalysisSettings object

        Parameters
        ----------
        analysis_settings : AnalysisSettings
            Analysis settings object
        parallel : bool, optional
            Whether to run the channels in parallel for `binaural.mosqito_metric_2ch`, by default True
            If False, will run each channel sequentially.
            If being run as part of a larger parallel analysis (e.g. processing many recordings at once), this will
            automatically be set to False.
            Applies only to `binaural.mosqito_metric_2ch`. The other metrics are considered fast enough not to bother.
        verbose : bool, optional
            Whether to print status updates, by default False

        Returns
        -------
        pd.DataFrame
            MultiIndex Dataframe of results.
            Index includes "Recording" and "Channel" with a column for each metric.
        """
        return process_all_metrics(self, analysis_settings, parallel, verbose)


__all__ = ["Binaural"]



import json
import time
from pathlib import Path
from tqdm.auto import tqdm

from soundscapy import AnalysisSettings
from soundscapy import Binaural
from soundscapy.analysis.binaural import *


def load_analyse_binaural(wav_file, levels, analysis_settings, verbose=True):
    """Load and analyse binaural file

    Parameters
    ----------
    wav_file : str
        Path to wav file
    levels : list
        List of levels to analyse
    analysis_settings : AnalysisSettings
        Analysis settings
    verbose : bool, optional
        Print progress, by default True

    Returns
    -------
    results : dict
        Dictionary with results
    """
    print(f"Processing {wav_file.stem}")
    b = Binaural.from_wav(wav_file)
    decibel = (levels[b.recording]["Left"], levels[b.recording]["Right"])
    b.calibrate_to(decibel, inplace=True)
    return process_all_metrics(b, analysis_settings, parallel=False, verbose=verbose)


def parallel_process(wav_files, results_df, levels, analysis_settings, verbose=True):
    """
    Parallel processing of binaural files

    Parameters
    ----------
    wav_files : list
        List of wav files
    results_df : pandas.DataFrame
        Results dataframe
    levels : dict
        Dictionary with levels
    analysis_settings : AnalysisSettings
        Analysis settings
    verbose : bool, optional
        Print progress, by default True

    Returns
    -------
    results_df : pandas.DataFrame
        Results dataframe
    """
    # Parallel processing with Pool.apply_async() without callback function

    pool = mp.Pool(mp.cpu_count() - 1)
    results = []
    result_objects = [
        pool.apply_async(
            load_analyse_binaural,
            args=(wav_file, levels, analysis_settings, verbose),
        )
        for wav_file in wav_files
    ]
    with tqdm(total = len(result_objects), desc="Processing files") as pbar:
        for r in result_objects:
            r.wait()
            results.append(r.get())
            pbar.update()
    # results = [r.get() for r in result_objects]

    pool.close()
    pool.join()

    for r in results:
        results_df = add_results(results_df, r)

    return results_df


if __name__ == "__main__":
    from datetime import datetime

    base_path = Path()
    wav_folder = base_path.joinpath("test", "data")
    levels = wav_folder.joinpath("Levels.json")
    with open(levels) as f:
        levels = json.load(f)

    analysis_settings = AnalysisSettings.default()

    df = prep_multiindex_df(levels, incl_metric=True)

    start = time.perf_counter()

    df = parallel_process(
        wav_folder.glob("*.wav"), df, levels, analysis_settings, verbose=True
    )

    end = time.perf_counter()
    print(f"Time taken: {end - start:.2f} seconds")

    df.to_excel(base_path.joinpath("test", f"ParallelTest_{datetime.today()}.xlsx"))



import multiprocessing as mp

import pandas as pd

from soundscapy.analysis.metrics import *

DEFAULT_LABELS = {
    "LZeq": "LZeq",
    "LCeq": "LCeq",
    "LAeq": "LAeq",
    "Leq": "Leq",
    "SEL": "SEL",
    "loudness_zwtv": "N",
    "roughness_dw": "R",
    "sharpness_din_perseg": "S",
    "sharpness_din_from_loudness": "S",
    "sharpness_din_tv": "S",
}

# 2ch Metrics calculations


def pyacoustics_metric_2ch(
    b,
    metric: str,
    statistics: Union[tuple, list] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Union[tuple, list] = ("Left", "Right"),
    as_df: bool = False,
    return_time_series: bool = False,
    verbose: bool = False,
    func_args={},
):
    """Run a metric from the python acoustics library on a Binaural object.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the metric for
    metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
        The metric to run
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc),
            by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
    label : str, optional
        Label to use for the metric in the results dictionary, by default None
        If None, will pull from default label for that metric given in DEFAULT_LABELS
    channel_names : tuple or list, optional
        Custom names for the channels, by default ("Left", "Right")
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False
        Cannot return time series if as_df is True
    verbose : bool, optional
        Whether to print status updates, by default False
    func_args : dict, optional
        Arguments to pass to the metric function, by default {}

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame

    See Also
    --------
    sq_metrics.pyacoustics_metric_1ch
    """
    if b.channels != 2:
        raise ValueError(
            "Must be 2 channel signal. Use `pyacoustics_metric_1ch instead`."
        )

    if verbose:
        print(f" - Calculating Python Acoustics metrics: {metric}")
    res_l = pyacoustics_metric_1ch(
        b[0],
        metric,
        statistics,
        label,
        as_df=False,
        return_time_series=return_time_series,
        func_args=func_args,
    )

    res_r = pyacoustics_metric_1ch(
        b[1],
        metric,
        statistics,
        label,
        as_df=False,
        return_time_series=return_time_series,
        func_args=func_args,
    )

    res = {channel_names[0]: res_l, channel_names[1]: res_r}
    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


def _parallel_mosqito_metric_2ch(
    b,
    metric: str,
    statistics: Union[tuple, list] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Union[tuple, list] = ("Left", "Right"),
    return_time_series: bool = False,
    func_args={},
):
    """Run a metric from the mosqito library on a Binaural object.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the metric for
    metric : {"LZeq", "Leq", "LAeq", "LCeq", "SEL"}
        The metric to run
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc),
            by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
    label : str, optional
        Label to use for the metric in the results dictionary, by default None
        If None, will pull from default label for that metric given in DEFAULT_LABELS
    channel_names : tuple or list, optional
        Custom names for the channels, by default ("Left", "Right")
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False
        Cannot return time series if as_df is True
    func_args : dict, optional
        Arguments to pass to the metric function, by default {}

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame

    See Also
    --------
    sq_metrics.mosqito_metric_1ch
    """
    pool = mp.Pool(mp.cpu_count())
    result_objects = pool.starmap(
        mosqito_metric_1ch,
        [
            (
                b[i],
                metric,
                statistics,
                label,
                False,
                return_time_series,
                func_args,
            )
            for i in [0, 1]
        ],
    )

    pool.close()
    return {channel: result_objects[i] for i, channel in enumerate(channel_names)}


def mosqito_metric_2ch(
    b,
    metric: str,
    statistics: Union[tuple, list] = (
        5,
        10,
        50,
        90,
        95,
        "avg",
        "max",
        "min",
        "kurt",
        "skew",
    ),
    label: str = None,
    channel_names: Union[tuple, list] = ("Left", "Right"),
    as_df: bool = False,
    return_time_series: bool = False,
    parallel: bool = True,
    verbose: bool = False,
    func_args={},
):
    """function for calculating metrics from Mosqito.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the sound quality indices for
    metric : {"loudness_zwtv", "sharpness_din_from_loudness", "sharpness_din_perseg",
    "sharpness_din_tv", "roughness_dw"}
        Metric to calculate
    statistics : tuple or list, optional
        List of level statistics to calculate (e.g. L_5, L_90, etc.),
            by default (5, 10, 50, 90, 95, "avg", "max", "min", "kurt", "skew")
    label : str, optional
        Label to use for the metric in the results dictionary, by default None
        If None, will pull from default label for that metric given in DEFAULT_LABELS
    channel_names : tuple or list, optional
        Custom names for the channels, by default ("Left", "Right")
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    return_time_series : bool, optional
        Whether to return the time series of the metric, by default False
        Only works for metrics that return a time series array.
        Cannot be returned in a dataframe. Will raise a warning if both `as_df`
        and `return_time_series` are True and will only return the DataFrame with the other stats.
    parallel : bool, optional
        Whether to run the channels in parallel, by default True
        If False, will run each channel sequentially.
        If being run as part of a larger parallel analysis (e.g. processing many recordings at once), this will
        automatically be set to False.
    verbose : bool, optional
        Whether to print status updates, by default False
    func_args : dict, optional
        Additional arguments to pass to the metric function, by default {}

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame
    """
    if b.channels != 2:
        raise ValueError("Must be 2 channel signal. Use `mosqito_metric_1ch` instead.")
    if verbose:
        if metric == "sharpness_din_from_loudness":
            print(
                " - Calculating MoSQITo metrics: `sharpness_din` from `loudness_zwtv`"
            )
        else:
            print(f" - Calculating MoSQITo metric: {metric}")

    # Make sure we're not already running in a parallel process
    # (e.g. if called from `parallel_process`)
    if mp.current_process().daemon:
        parallel = False
    if parallel:
        res = _parallel_mosqito_metric_2ch(
            b, metric, statistics, label, channel_names, return_time_series
        )

    else:
        res_l = mosqito_metric_1ch(
            b[0],
            metric,
            statistics,
            label,
            as_df=False,
            return_time_series=return_time_series,
            func_args=func_args,
        )

        res_r = mosqito_metric_1ch(
            b[1],
            metric,
            statistics,
            label,
            as_df=False,
            return_time_series=return_time_series,
            func_args=func_args,
        )

        res = {channel_names[0]: res_l, channel_names[1]: res_r}
    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


def maad_metric_2ch(
    b,
    metric: str,
    channel_names: Union[tuple, list] = ("Left", "Right"),
    as_df: bool = False,
    verbose: bool = False,
    func_args={},
):
    """Run a metric from the scikit-maad library (or suite of indices) on a binaural signal.

    Currently only supports running all the alpha indices at once.

    Parameters
    ----------
    b : Binaural
        Binaural signal to calculate the alpha indices for
    metric : {"all_temporal_alpha_indices", "all_spectral_alpha_indices"}
        Metric to calculate
    channel_names : tuple or list, optional
        Custom names for the channels, by default ("Left", "Right").
    as_df : bool, optional
        Whether to return a pandas DataFrame, by default False
        If True, returns a MultiIndex Dataframe with ("Recording", "Channel") as the index.
    verbose : bool, optional
        Whether to print status updates, by default False
    func_args : dict, optional
        Additional arguments to pass to the metric function, by default {}

    Returns
    -------
    dict or pd.DataFrame
        Dictionary of results if as_df is False, otherwise a pandas DataFrame

    See Also
    --------
    scikit-maad library
    `sq_metrics.maad_metric_1ch`

    """
    if b.channels != 2:
        raise ValueError("Must be 2 channel signal. Use `maad_metric_1ch` instead.")
    if verbose:
        print(f" - Calculating scikit-maad {metric}")
    res_l = maad_metric_1ch(b[0], metric, as_df=False)
    res_r = maad_metric_1ch(b[1], metric, as_df=False)
    res = {channel_names[0]: res_l, channel_names[1]: res_r}
    if not as_df:
        return res
    try:
        rec = b.recording
    except AttributeError:
        rec = 0
    df = pd.DataFrame.from_dict(res, orient="index")
    df["Recording"] = rec
    df["Channel"] = df.index
    df.set_index(["Recording", "Channel"], inplace=True)
    return df


# Analysis dataframe functions
def prep_multiindex_df(dictionary: dict, label: str = "Leq", incl_metric: bool = True):
    """df help to prepare a MultiIndex dataframe from a dictionary of results

    Parameters
    ----------
    dictionary : dict
        Dict of results with recording name as key, channels {"Left", "Right"} as second key, and Leq metric as value
    label : str, optional
        Name of metric included, by default "Leq"
    incl_metric : bool, optional
        Whether to include the metric value in the resulting dataframe, by default True
        If False, will only set up the DataFrame with the proper MultiIndex
    Returns
    -------
    pd.DataFrame
        Index includes "Recording" and "Channel" with a column for each index if `incl_metric`.

    """
    new_dict = {}
    for outerKey, innerDict in dictionary.items():
        for innerKey, values in innerDict.items():
            new_dict[(outerKey, innerKey)] = values
    idx = pd.MultiIndex.from_tuples(new_dict.keys())
    df = pd.DataFrame(new_dict.values(), index=idx, columns=[label])
    df.index.names = ["Recording", "Channel"]
    if not incl_metric:
        df = df.drop(columns=[label])
    return df


def add_results(results_df: pd.DataFrame, metric_results: pd.DataFrame):
    """Add results to MultiIndex dataframe

    Parameters
    ----------
    results_df : pd.DataFrame
        MultiIndex dataframe to add results to
    metric_results : pd.DataFrame
        MultiIndex dataframe of results to add

    Returns
    -------
    pd.DataFrame
        Index includes "Recording" and "Channel" with a column for each index.
    """
    if not set(metric_results.columns).issubset(set(results_df.columns)):
        # Check if results_df already has the columns in results
        results_df = results_df.join(metric_results)
    else:
        results_df.update(metric_results, errors="ignore")
    return results_df


def process_all_metrics(
    b,
    analysis_settings,
    parallel: bool = True,
    verbose: bool = False,
):
    """Loop through all metrics included in `analysis_settings` and add results to `results_df`

    Parameters
    ----------
    b : Binaural
        Binaural signal to process
    analysis_settings : AnalysisSettings
        Settings for analysis, including `run` tag for whether to run a metric
    parallel : bool, optional
        Whether to run the channels in parallel for `binaural.mosqito_metric_2ch`, by default True
        If False, will run each channel sequentially.
        If being run as part of a larger parallel analysis (e.g. processing many recordings at once), this will
        automatically be set to False.
        Applies only to `binaural.mosqito_metric_2ch`. The other metrics are considered fast enough not to bother.
    verbose : bool, optional
        Whether to print status updates, by default False

    Returns
    -------
    pd.DataFrame
        MultiIndex DataFrame with results from all metrics for one Binaural recording
    """
    if verbose:
        print(f"Processing {b.recording}")

    idx = pd.MultiIndex.from_tuples(((b.recording, "Left"), (b.recording, "Right")))
    results_df = pd.DataFrame(index=idx)
    results_df.index.names = ["Recording", "Channel"]
    
    # Count number of metrics to run
    metric_count = 0
    for library in analysis_settings.keys():
        if library not in ["PythonAcoustics", "scikit-maad", "MoSQITo"]:
            pass
        else:
            for metric in analysis_settings[library].keys():
                if analysis_settings[library][metric]["run"]:
                    metric_count += 1

    # Loop through options in analysis_settings
    for library in analysis_settings.keys():
        # Python Acoustics metrics
        if library == "PythonAcoustics":
            for metric in analysis_settings[library].keys():
                results_df = pd.concat(
                    (
                        results_df,
                        b.pyacoustics_metric(
                            metric, verbose=verbose, analysis_settings=analysis_settings
                        ),
                    ),
                    axis=1,
                )
        # MoSQITO metrics
        elif library == "MoSQITo":
            for metric in analysis_settings[library].keys():
                results_df = pd.concat(
                    (
                        results_df,
                        b.mosqito_metric(
                            metric,
                            parallel=parallel,
                            verbose=verbose,
                            analysis_settings=analysis_settings,
                        ),
                    ),
                    axis=1,
                )
        # scikit-maad metrics
        elif library == "scikit-maad":
            for metric in analysis_settings[library].keys():
                results_df = pd.concat(
                    (
                        results_df,
                        b.maad_metric(
                            metric, verbose=verbose, analysis_settings=analysis_settings
                        ),
                    ),
                    axis=1,
                )

    return results_df


# %%



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['soundscapy',
 'soundscapy.analysis',
 'soundscapy.databases',
 'soundscapy.plotting',
 'soundscapy.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'acoustics>=0.2.6,<0.3.0',
 'mosqito>=1.0.8,<2.0.0',
 'openpyxl>=3.0.10,<4.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pyjanitor==0.23.1',
 'pytest>=7.2.2,<8.0.0',
 'scikit-maad>=1.3.12,<2.0.0',
 'seaborn>=0.12.2,<0.13.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'soundscapy',
    'version': '0.4.3',
    'description': 'A python library for analysing and visualising soundscape assessments.',
    'long_description': '<img src="https://raw.githubusercontent.com/MitchellAcoustics/Soundscapy/main/docs/logo/LightLogo.png" width="300">\n\n# Soundscapy\n\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/MitchellAcoustics/Soundscapy/main?labpath=docs%2Ftutorials%2FHowToAnalyseAndRepresentSoundscapes.ipynb)\n\nA python library for analysing and visualising soundscape assessments. \n\n**Disclaimer:** This module is still heavily in development, and might break what you\'re working on. It will also likely require a decent amount of troubleshooting at this stage. I promise bug fixes and cleaning up is coming!\n\n## Installation\n\nThe package is still under development, but can be installed with pip:\n\n```\npip install soundscapy\n```\n\n## Examples\n\nWe are currently working on writing more comprehensive examples and documentation, please bear with us in the meantime. \n\nAn example notebook which reproduces the figures used in our recent paper "How to analyse and represent quantitative soundscape data" is provided in the `examples` folder.\n\n## Acknowledgements\n\nThe newly added Binaural analysis functionality relies directly on three acoustic analysis libraries: \n * [python-acoustics](https://github.com/python-acoustics/python-acoustics) for the standard environmental and building acoustics metrics, \n * [scikit-maad](https://github.com/scikit-maad/scikit-maad) for the bioacoustics and ecological soundscape metrics, and\n * [MoSQITo](https://github.com/Eomys/MoSQITo) for the psychoacoustics metrics. We thank each of these packages for their great work in making advanced acoustic analysis more accessible.\n\n## Citation\n\nIf you are using Soundscapy in your research, please help our scientific visibility by citing our work! Please include a citation to our accompanying paper:\n\nMitchell, A., Aletta, F., & Kang, J. (2022). How to analyse and represent quantitative soundscape data. *JASA Express Letters, 2*, 37201. [https://doi.org/10.1121/10.0009794](https://doi.org/10.1121/10.0009794)\n\n\n<!---\nBibtex:\n```\n@Article{Mitchell2022How,\n  author         = {Mitchell, Andrew and Aletta, Francesco and Kang, Jian},\n  journal        = {JASA Express Letters},\n  title          = {How to analyse and represent quantitative soundscape data},\n  year           = {2022},\n  number         = {3},\n  pages          = {037201},\n  volume         = {2},\n  doi            = {10.1121/10.0009794},\n  eprint         = {https://doi.org/10.1121/10.0009794},\n}\n\n```\n--->\n\n## Development Plans\n\nAs noted, this package is in heavy development to make it more useable, more stable, and to add features and improvements. At this stage it is mostly limited to doing basic quality checks of soundscape survey data and creating the soundscape distribution plots. Some planned improvements are:\n\n - [ ] Simplify the plotting options\n - [ ] Possibly improve how the plots and data are handled - a more OOP approach would be good.\n - [ ] Add appropriate tests and documentation.\n - [ ] Bug fixes, ~~particularly around setting color palettes.~~\n\nLarge planned feature additions are:\n\n - [ ] Add better methods for cleaning datasets, including robust outlier exclusion and imputation.\n - [x] Add handling of .wav files.\n - [x] Integrate environmental acoustic and psychoacoustic batch processing. This will involve using existing packages (e.g. MoSQito, python-acoustics) to do the metric calculations, but adding useful functionality for processing many files at once, tieing them to a specific survey response, and implementing a configuration file for maintaining consistent analysis settings.\n - [ ] Integrate the predictive modelling results from the SSID team\'s research to enable a single pipelined from recording -> psychoacoustics -> predicted soundscape perception (this is very much a pie-in-the-sky future plan, which may not be possible).\n\n### Contributing\n\nIf you would like to contribute or if you have any bugs you have found while using `Soundscapy\', please feel free to get in touch or submit an issue or pull request!\n',
    'author': 'Andrew Mitchell',
    'author_email': 'andrew.mitchell.18@ucl.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MitchellAcoustics/Soundscapy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
